# cspell:ignore KEY_NPAGE, KEY_PPAGE
# pylint: disable=too-many-lines
"""The main UI renderer."""

from __future__ import annotations

import curses
import logging
import re

from collections.abc import Callable
from collections.abc import Mapping
from collections.abc import Sequence
from curses import ascii as curses_ascii
from functools import cache
from math import ceil
from math import floor
from re import Match
from re import Pattern
from typing import TYPE_CHECKING
from typing import Any
from typing import NamedTuple
from typing import Protocol

from ansible_navigator.content_defs import ContentFormat
from ansible_navigator.content_defs import ContentType
from ansible_navigator.content_defs import ContentTypeSequence
from ansible_navigator.content_defs import ContentView
from ansible_navigator.utils.functions import templar
from ansible_navigator.utils.serialize import serialize

from .colorize import Colorize
from .colorize import rgb_to_ansi
from .curses_defs import CursesLine
from .curses_defs import CursesLinePart
from .curses_defs import CursesLines
from .curses_defs import RgbTuple
from .curses_defs import SimpleLinePart
from .curses_window import CursesWindow
from .curses_window import Window
from .field_text import FieldText
from .form_handler_text import FormHandlerText
from .form_utils import warning_notification
from .menu_builder import MenuBuilder
from .ui_constants import Decoration


if TYPE_CHECKING:
    from .form import Form
    from .ui_config import UIConfig


KEY_REFRESH = "KEY_F(5)"

STANDARD_KEYS = {
    "^b/PgUp": "page up",
    "^f/PgDn": "page down",
    "\u2191\u2193": "scroll",
    "esc": "back",
}
END_KEYS = {":help": "help"}


class Action(NamedTuple):
    """the user's input."""

    value: str | int
    match: Match[str]


class Content(NamedTuple):
    """what's on the screen, when showing content."""

    showing: Any


class Menu(NamedTuple):
    """details about the currently showing menu."""

    current: ContentTypeSequence
    columns: list[str]


class ContentFormatCallable(Protocol):
    """Protocol definition for the Ui.content_format callable."""

    def __call__(
        self,
        value: ContentFormat | None = None,
        default: bool = False,
    ) -> ContentFormat:
        """Refer to and keep in sync with UserInterface.content_format.

        Args:
            value: The value refers to the UI content format
            default: The default UI content format
        """


class ShowCallable(Protocol):
    """Protocol definition for the Ui.show callable."""

    # pylint: disable=too-many-arguments
    def __call__(
        self,
        obj: ContentType,
        content_format: ContentFormat | None = None,
        index: int | None = None,
        columns: list[str] | None = None,
        await_input: bool = True,
        filter_content_keys: Callable[..., Any] = lambda x: x,
        color_menu_item: Callable[..., Any] = lambda *args, **kwargs: (0, 0),
        content_heading: Callable[..., Any] = lambda *args, **kwargs: None,
    ) -> Interaction:
        """Refer to and keep in sync with UserInterface.show.

        Args:
            obj: The inbound object
            content_format: Refers to the content format
            index: The indices of objects
            columns: Refers to the menu column
            await_input: Should we wait for user input?
            filter_content_keys: The filter content keys
            color_menu_item: To color the menu item
            content_heading: Refers to the content heading
        """


class Ui(NamedTuple):
    """select functions that can be called from an action."""

    clear: Callable[..., Any]
    menu_filter: Callable[..., Any]
    scroll: Callable[..., Any]
    show: ShowCallable
    show_form: Callable[[Form], Form]
    update_status: Callable[..., Any]
    content_format: ContentFormatCallable


class Interaction(NamedTuple):
    """wrapper for what is sent back to the calling app."""

    name: str
    action: Action
    ui: Ui
    content: Content | None = None
    menu: Menu | None = None


class UserInterface(CursesWindow):
    # pylint: disable=too-many-instance-attributes
    # pylint: disable=too-many-arguments
    """The main UI class."""

    def __init__(
        self,
        screen_min_height: int,
        kegexes: Callable[..., Any],
        refresh: int,
        ui_config: UIConfig,
        progress_bar_width: int = 8,
        status_width: int = 12,
    ) -> None:
        """Initialize the user interface.

        Args:
            screen_min_height: The minimum screen height
            kegexes: A callable producing a list of action regular
                expressions to match against
            refresh: The screen refresh time is ms
            ui_config: The current UI configuration
            progress_bar_width: The width of the progress bar
            status_width: The width of the status indicator
        """
        super().__init__(ui_config=ui_config)
        self._color_menu_item: Callable[[int, str, dict[str, Any]], tuple[int, int]]
        self._colorizer = Colorize(
            grammar_dir=self._ui_config.grammar_dir,
            theme_path=self._ui_config.theme_path,
        )
        self._content_heading: Callable[[Any, int], CursesLines | None]
        self._default_colors = None
        self._default_pairs = None
        self._default_content_format = ContentFormat.YAML
        self._filter_content_keys: Callable[[Any], dict[Any, Any]]
        self._hide_keys = True
        self._kegexes = kegexes
        self._logger = logging.getLogger(__name__)
        self._menu_filter: Pattern[str] | None = None
        self._menu_indices: tuple[int, ...] = ()

        self._progress_bar_width = progress_bar_width
        self._status_width = status_width
        self._prefix_color = 8
        self._refresh = [refresh]
        self._rgb_to_curses_color_idx: dict[RgbTuple, int] = {}
        self._screen_min_height = screen_min_height
        self._scroll = 0
        self._content_format = self._default_content_format
        self._status = ""
        self._status_color = 0
        self._screen: Window = curses.initscr()
        self._screen.timeout(refresh)
        self._screen.keypad(True)  # Enable keypad mode
        self._one_line_input = FormHandlerText(screen=self._screen, ui_config=self._ui_config)
        # This tracks which visible row to highlight
        self._highlight_line_offset: int | None = None
        # Cursor position in menus; None means cursor is not yet active.
        # It activates (becomes an int) only after the user presses UP or DOWN.
        self._menu_cursor_pos: int | None = None
        # Tracks the identity of the current menu's data object so the cursor
        # is reset only when entering a new menu, not on refresh cycles.
        self._menu_current_id: int | None = None

    def clear(self) -> None:
        """Clear the screen."""
        self._screen.clear()
        self._screen.refresh()

    def disable_refresh(self) -> None:
        """Disable the screen refresh."""
        self._refresh.append(self._refresh[-1])
        self._refresh.append(-1)
        self._screen.timeout(-1)

    def restore_refresh(self) -> None:
        """Restore the screen refresh to the previous value."""
        self._refresh.pop()
        self._screen.timeout(self._refresh.pop())

    def update_status(self, status: str = "", status_color: int = 0) -> None:
        """Update the status.

        Args:
            status: The string of status information
            status_color: The color of status
        """
        self._status = status
        self._status_color = status_color

    def menu_filter(self, value: str | None = "") -> Pattern[str] | None:
        """Set or return the menu filter.

        Args:
            value: None or the menu_filter regex to set

        Returns:
            the current menu filter
        """
        if value != "":
            if value is None:
                self._menu_filter = None
            else:
                try:
                    self._menu_filter = re.compile(value)
                except re.error:
                    self._menu_filter = None
                    self._logger.exception("Regex for menu filter was invalid: %s", value)
        return self._menu_filter

    def scroll(self, value: int | None = None) -> int:
        """Set or return the current scroll.

        Args:
            value: the value to set the scroll to

        Raises:
            TypeError: raise exception here

        Returns:
            the current scroll
        """
        if value is not None:
            if not isinstance(value, int):
                raise TypeError
            self._scroll = value
        return self._scroll

    def content_format(
        self,
        value: ContentFormat | None = None,
        default: bool = False,
    ) -> ContentFormat:
        """Set or return the current content format.

        Args:
            value: The value to set the content format to
            default: The default content format

        Returns:
            The current content format
        """
        if value is not None:
            self._content_format = value
            if default:
                self._default_content_format = value
        return self._content_format

    @property
    def _ui(self) -> Ui:
        """Limit the callables the actions can access.

        Returns:
            A tuple of available functions
        """
        res = Ui(
            clear=self.clear,
            menu_filter=self.menu_filter,
            scroll=self.scroll,
            show=self.show,
            show_form=self.show_form,
            update_status=self.update_status,
            content_format=self.content_format,
        )
        return res

    def _footer(self, key_dict: dict[Any, Any]) -> CursesLine:
        """Build a footer from the key dict spread the columns out evenly.

        Args:
            key_dict: the keys and their description

        Returns:
            The footer line
        """
        column_widths = [len(f"{k!s}: {v!s}") for k, v in key_dict.items()]
        status_width = self._progress_bar_width if self._status else 0
        gap = floor((self._screen_width - status_width - sum(column_widths)) / len(key_dict))
        adjusted_column_widths = [c + gap for c in column_widths]
        col_starts = [0]
        for idx, column_width in enumerate(adjusted_column_widths):
            col_starts.append(column_width + col_starts[idx])
        footer = []
        for idx, key in enumerate(key_dict):
            left = key[0 : adjusted_column_widths[idx]]
            right = f" {key_dict[key]}"
            right = right[0 : adjusted_column_widths[idx] - len(key)]
            footer.append(
                CursesLinePart(
                    column=col_starts[idx],
                    string=left,
                    color=0,
                    decoration=curses.A_BOLD,
                ),
            )
            footer.append(
                CursesLinePart(
                    column=col_starts[idx] + len(left),
                    string=right,
                    color=0,
                    decoration=0,
                ),
            )
        if self._status:
            # place the status to the far right -1 for the scrollbar
            # center place the uneven extra on the right, so flip it twice
            status = self._status.capitalize()
            status = status[0 : self._status_width - 1]  # max
            status = status[::-1]  # reverse
            status = status.center(self._status_width)  # pad
            status = status[::-1]  # reverse
            footer.append(
                CursesLinePart(
                    column=self._screen_width - self._status_width - 1,
                    string=status,
                    color=self._status_color,
                    decoration=curses.A_REVERSE,
                ),
            )
        return CursesLine(tuple(footer))

    def _scroll_bar(
        self,
        viewport_height: int,
        len_heading: int,
        menu_size: int,
        body_start: int,
        body_stop: int,
    ) -> None:
        """Add a scroll bar if the length of the content is longer than the viewport height.

        Args:
            viewport_height: The height of the viewport
            len_heading: The height of the heading
            menu_size: The number of lines in the content
            body_start: Where we are in the body
            body_stop: The end of the body
        """
        start_scroll_bar = body_start / menu_size * viewport_height
        stop_scroll_bar = body_stop / menu_size * viewport_height
        len_scroll_bar = ceil(stop_scroll_bar - start_scroll_bar)
        color = self._prefix_color
        for idx in range(int(start_scroll_bar), int(start_scroll_bar + len_scroll_bar)):
            lineno = idx + len_heading
            line_part = CursesLinePart(
                column=self._screen_width - 1,
                string="\u2592",
                color=color,
                decoration=0,
            )
            self._add_line(
                window=self._screen,
                lineno=min(lineno, viewport_height + len_heading),
                line=CursesLine(
                    (line_part,),
                ),
            )

    def _get_input_line(self) -> str:
        """Get one line of input from the user.

        Returns:
            the lines
        """
        self.disable_refresh()
        form_field = FieldText(name="one_line", prompt="")
        line_part = CursesLinePart(column=0, string=":", color=0, decoration=0)
        input_at = self._screen_height - 1  # screen y is zero based
        self._add_line(
            window=self._screen,
            lineno=input_at,
            line=CursesLine(
                (line_part,),
            ),
        )
        self._screen.refresh()
        self._one_line_input.win = curses.newwin(1, self._screen_width, input_at, 1)
        self._one_line_input.win.keypad(True)
        while True:
            user_input, char = self._one_line_input.handle(0, [form_field])
            if char == curses_ascii.ESC:
                break
            if char in (curses.KEY_ENTER, 10, 13):
                break
            if char == curses.KEY_RESIZE:
                break
        self.restore_refresh()
        self._curs_set(0)
        return user_input

    def _handle_display_key(
        self,
        key: str,
        keypad: set[str],
        other_valid_keys: list[str],
        viewport_height: int,
        heading_len: int,
        footer_len: int,
        count: int,
        lines_len: int,
        indent_heading: int,
    ) -> str | None:
        """Handle a key press during display, adjusting scroll as needed.

        Args:
            key: The key that was pressed
            keypad: The set of numeric keypad characters
            other_valid_keys: Other keys that should be returned as-is
            viewport_height: The height of the viewport
            heading_len: The height of the heading
            footer_len: The height of the footer
            count: The total number of lines
            lines_len: The number of currently visible lines
            indent_heading: The indentation of heading

        Returns:
            The key string to return, or None if the key was not handled
        """
        if key == "KEY_RESIZE":
            new_scroll = min(
                self._scroll - viewport_height + self._screen_height - heading_len - footer_len,
                lines_len,
            )
            self.scroll(new_scroll)
            return key
        if key in keypad or key in other_valid_keys:
            return key
        if key == "KEY_DOWN":
            if not indent_heading:
                self.scroll(max(min(self.scroll() + 1, count), viewport_height))
            return key
        if key == "KEY_UP":
            if not indent_heading:
                self.scroll(max(self.scroll() - 1, viewport_height))
            return key
        if key in ["^F", "KEY_NPAGE"]:
            self.scroll(max(min(self.scroll() + viewport_height, count), viewport_height))
            return key
        if key in ["^B", "KEY_PPAGE"]:
            self.scroll(max(self.scroll() - viewport_height, viewport_height))
            return key
        if key == ":":
            return ":"
        return None

    def _display(
        self,
        lines: CursesLines,
        line_numbers: tuple[int, ...],
        heading: CursesLines | None,
        indent_heading: int,
        key_dict: dict[Any, Any],
        await_input: bool,
        count: int,
    ) -> str:
        """Show something on the screen.

        Args:
            lines: The lines to show
            line_numbers: The number of lines to show
            heading: The headers to show
            indent_heading: The indentation of heading
            key_dict: any supplemental key to show
            await_input: Should we wait for a key
            count: The count to show

        Returns:
            the key pressed
        """
        heading = heading or CursesLines(())
        heading_len = len(heading)
        footer = self._footer({**STANDARD_KEYS, **key_dict, **END_KEYS})
        footer_at = self._screen_height - 1  # screen is 0 based index
        footer_len = 1

        viewport_height = self._screen_height - len(heading) - footer_len
        self.scroll(max(self.scroll(), viewport_height))

        index_width = len(str(count))

        keypad = {str(x) for x in range(10)}
        other_valid_keys = [
            "+",
            "-",
            "_",
            KEY_REFRESH,
            "^[",
            "\x1b",
            "CURSOR_ENTER",
        ]

        while True:
            self._screen.erase()
            prefix = " " * (index_width + len("|")) if indent_heading else None

            # Add the heading
            for idx, line in enumerate(heading):
                self._add_line(window=self._screen, lineno=idx, line=line, prefix=prefix)

            # Add the content
            for idx, line in enumerate(lines):
                line_index = line_numbers[idx]
                line_index_str = str(line_index).rjust(index_width)
                prefix = f"{line_index_str}\u2502"
                # Apply highlight decoration when this is the selected row
                if (
                    indent_heading
                    and self._highlight_line_offset is not None
                    and idx == self._highlight_line_offset
                ):
                    # Rebuild the line with reverse-video decoration added
                    highlighted_parts = tuple(
                        CursesLinePart(
                            column=lp.column,
                            string=lp.string,
                            color=lp.color,
                            decoration=lp.decoration | curses.A_REVERSE,
                        )
                        for lp in line
                    )
                    line_to_draw = CursesLine(highlighted_parts)
                else:
                    line_to_draw = line
                self._add_line(
                    window=self._screen,
                    lineno=idx + len(heading),
                    line=line_to_draw,
                    prefix=prefix,
                )

            # Add the scroll bar
            if count > viewport_height:
                self._scroll_bar(
                    viewport_height=viewport_height,
                    len_heading=len(heading),
                    menu_size=count,
                    body_start=self._scroll - viewport_height,
                    body_stop=self._scroll,
                )

            # Add the footer after the rest of the screen has been drawn
            self._add_line(window=self._screen, lineno=footer_at, line=footer)

            self._screen.refresh()

            if await_input:
                char = self._screen.getch()
                key = KEY_REFRESH if char == -1 else curses.keyname(char).decode()
                if char in [10, 13]:  # Enter key codes: 10=LF, 13=CR
                    key = "CURSOR_ENTER"
            else:
                key = KEY_REFRESH

            return_value = self._handle_display_key(
                key=key,
                keypad=keypad,
                other_valid_keys=other_valid_keys,
                viewport_height=viewport_height,
                heading_len=heading_len,
                footer_len=footer_len,
                count=count,
                lines_len=len(lines),
                indent_heading=indent_heading,
            )
            if return_value == ":":
                colon_entry = self._get_input_line()
                if colon_entry is None:
                    continue
                return_value = colon_entry

            if return_value is not None:
                return return_value

    def _template_match_action(
        self,
        entry: str,
        current: Any,
    ) -> tuple[str, Action] | tuple[None, None]:
        """Attempt to template & match the user input against the kegexes.

        Args:
            entry: the user input
            current: the content on the screen

        Returns:
            The name of the action and the action to call or nothing if
            no match found
        """
        if (
            not entry.startswith("{{") and "{{" in entry and "}}" in entry
        ):  # don't match pure template
            if isinstance(current, Mapping):
                template_vars = current
                type_msgs = []
            else:
                template_vars = {"this": current}
                type_msgs = ["Current content passed for templating is not a dictionary."]
                type_msgs.append("[HINT] Use 'this' to reference it (e.g. {{ this[0] }}")
            errors, entry = templar(entry, template_vars)
            if errors:
                msgs = ["Errors encountered while templating input"] + errors
                msgs.extend(type_msgs)
                self._show_form(warning_notification(msgs))
                return None, None
        for kegex in self._kegexes():
            match = kegex.kegex.match(entry)
            if match:
                return kegex.name, Action(match=match, value=entry)

        msgs = [f"Could not find a match for ':{entry}'"]
        msgs.append("[Hint] Try ':help' for a list of available commands.")
        self._show_form(warning_notification(msgs))
        return None, None

    def _serialize_color(self, obj: Any) -> CursesLines:
        """Serialize, if necessary and color an obj.

        Args:
            obj: the object to color

        Returns:
            The generated lines
        """
        if self.content_format() is ContentFormat.ANSI:
            return self._colorizer.render_ansi(doc=obj)

        content_view = ContentView.NORMAL if self._hide_keys else ContentView.FULL
        current_format = self.content_format()
        if current_format.value.serialization:
            string = serialize(
                content_view=content_view,
                content=obj,
                serialization_format=current_format.value.serialization,
            )
        else:
            string = obj

        scope = "no_color"
        if self._ui_config.color:
            scope = self.content_format().value.scope

        rendered = self._colorizer.render(doc=string, scope=scope)
        self._cache_init_colors(rendered)
        return self._color_decorate_lines(rendered)

    def _cache_init_colors(self, lines: list[list[SimpleLinePart]]) -> None:
        """Cache and init the unique colors for future use.

        Maintain a mapping of RGB colors
        to curses colors in self._rgb_to_curses_color_idx

        Args:
            lines: The from which colors will be cached and initialized
        """
        if curses.COLORS > 16 and self._term_osc4_support:
            unique_colors = list(
                {chars.color for line in lines for chars in line if chars.color},
            )
            # start custom colors at 16
            for color in unique_colors:
                scale = 1000 / 255
                red, green, blue = color
                if color not in self._rgb_to_curses_color_idx:
                    if not self._rgb_to_curses_color_idx:
                        curses_colors_idx = 16
                    else:
                        curses_colors_idx = max(self._rgb_to_curses_color_idx.values()) + 1

                    self._rgb_to_curses_color_idx[color] = curses_colors_idx
                    curses.init_color(
                        curses_colors_idx,
                        int(red * scale),
                        int(green * scale),
                        int(blue * scale),
                    )
                    self._logger.debug(
                        "Added color: %s:%s",
                        curses_colors_idx,
                        curses.color_content(curses_colors_idx),
                    )
                    curses.init_pair(curses_colors_idx, curses_colors_idx, -1)

    def _color_decorate_lines(self, lines: list[list[SimpleLinePart]]) -> CursesLines:
        """Color and decorate each of the lines.

        Args:
            lines: The lines to transform

        Returns:
            All lines colored
        """
        return CursesLines(tuple(self._color_decorate_line(line) for line in lines))

    def _color_decorate_line(self, line: list[SimpleLinePart]) -> CursesLine:
        """Color and decorate one line.

        Args:
            line: The line to color

        Returns:
            One line colored
        """
        return CursesLine(tuple(self._color_decorate_line_part(line_part) for line_part in line))

    def _color_decorate_line_part(self, line_part: SimpleLinePart) -> CursesLinePart:
        """Color and decorate one line part.

        Args:
            line_part: One line part

        Returns:
            One line part colored
        """
        if line_part.color:
            if self._term_osc4_support and curses.COLORS > 16:
                color = self._rgb_to_curses_color_idx[line_part.color]
            else:
                red, green, blue = line_part.color
                color = rgb_to_ansi(red, green, blue, curses.COLORS)
        else:
            color = 0

        decoration = Decoration.get_best(line_part.style)
        return CursesLinePart(
            column=line_part.column,
            string=line_part.chars,
            color=color,
            decoration=decoration,
        )

    def _filter_and_serialize(self, obj: Any) -> tuple[CursesLines | None, CursesLines]:
        """Filter an obj and serialize.

        Args:
            obj: the obj to serialize

        Returns:
            the serialize lines ready for display
        """
        heading = self._content_heading(obj, self._screen_width)
        filtered_obj = self._filter_content_keys(obj) if self._hide_keys else obj
        lines = self._serialize_color(filtered_obj)
        return heading, lines

    def _show_form(self, obj: Form) -> Form:
        """Show a form on the display.

        Args:
            obj: The form to show

        Returns:
            The form
        """
        res = obj.present(screen=self._screen, ui_config=self._ui_config)
        return res

    def _handle_obj_navigation(
        self,
        entry: str,
        objs: ContentTypeSequence,
        index: int,
    ) -> tuple[bool, int, str | None]:
        """Handle navigation keys when showing an object from a list.

        Args:
            entry: The key or command string entered by the user
            objs: The list of objects being navigated
            index: The current index into objs

        Returns:
            A tuple of (should_continue, new_index, new_entry) where
            should_continue indicates the caller should loop again,
            new_index is the possibly updated index, and new_entry
            is a signal string ("_" for toggle keys, "KEY_RESIZE"
            for resize, or None otherwise).
        """
        if entry in ["KEY_DOWN", "KEY_UP", "KEY_NPAGE", "KEY_PPAGE", "^F", "^B"]:
            return True, index, None

        if entry == "CURSOR_ENTER":
            return True, index, None

        if entry == "KEY_RESIZE":
            return True, index, "KEY_RESIZE"

        if entry == "_":
            return True, index, "_"

        if entry == "-":
            less = list(reversed([i for i in self._menu_indices if i - index < 0]))
            more = list(reversed([i for i in self._menu_indices if i - index > 0]))
            ordered_indices = less + more
            if ordered_indices:
                return True, ordered_indices[0], None
            return True, index, None

        if entry == "+":
            more = [i for i in self._menu_indices if i - index > 0]
            less = [i for i in self._menu_indices if i - index < 0]
            ordered_indices = more + less
            if ordered_indices:
                return True, ordered_indices[0], None
            return True, index, None

        if entry.isnumeric():
            return True, int(entry) % len(objs), None

        return False, index, None

    def _show_obj_from_list(
        self,
        objs: ContentTypeSequence,
        index: int,
        await_input: bool,
    ) -> Interaction:
        """Show an object on the display.

        Args:
            objs: A list of one or more object
            index: The index associated with an object
            await_input: Should we wait for user input before returning

        Returns:
            interaction with the user
        """
        heading, lines = self._filter_and_serialize(objs[index])
        while True:
            heading_len = len(heading) if heading is not None else 0
            footer_len = 1

            if self.scroll() == 0:
                last_line_idx = min(
                    len(lines) - 1,
                    self._screen_height - heading_len - footer_len - 1,
                )
            else:
                last_line_idx = min(len(lines) - 1, self._scroll - 1)

            first_line_idx = max(
                0,
                last_line_idx - (self._screen_height - 1 - heading_len - footer_len),
            )

            key_dict = {"-": "previous", "+": "next", "[0-9]": "goto"} if len(objs) > 1 else {}

            line_numbers = tuple(range(first_line_idx, last_line_idx + 1))

            entry = self._display(
                lines=CursesLines(lines[first_line_idx : last_line_idx + 1]),
                line_numbers=line_numbers,
                heading=heading,
                indent_heading=False,
                key_dict=key_dict,
                await_input=await_input,
                count=len(lines),
            )
            should_continue, new_index, new_entry = self._handle_obj_navigation(
                entry,
                objs,
                index,
            )
            if should_continue:
                if new_index != index:
                    index = new_index
                self.scroll(0)
                if new_entry == "_":
                    self._hide_keys = not self._hide_keys
                    heading, lines = self._filter_and_serialize(objs[index])
                elif new_entry == "KEY_RESIZE":
                    heading = self._content_heading(objs[index], self._screen_width)
                continue

            current = objs[index % len(objs)]

            name, action = self._template_match_action(entry, current)
            if name and action:
                if name == "refresh":
                    action = action._replace(value=index)

                filtered = self._filter_content_keys(current) if self._hide_keys else current
                content = Content(showing=filtered)
                return Interaction(name=name, action=action, content=content, ui=self._ui)

    def _obj_match_filter(self, obj: dict[Any, Any], columns: list[str]) -> bool:
        """Check columns in a dictionary against a regex.

        Args:
            obj: The dict to check
            columns: The dicts keys to check

        Returns:
            True if a match else False
        """
        return any(self._search_value(self.menu_filter(), obj.get(key)) for key in columns)

    @staticmethod
    @cache
    def _search_value(regex: Pattern[str], value: str) -> Match[str] | None:
        """Check a str against a regex.

        Args:
            regex: the compiled regex
            value: the string to check

        Returns:
            the match if made
        """
        return regex.search(str(value))

    def _get_heading_menu_items(
        self,
        current: Sequence[Any],
        columns: list[str],
        indices: tuple[int, ...],
    ) -> tuple[CursesLines, CursesLines]:
        """Build the menu.

        Args:
            current: A dict
            columns: The keys from the dictionary to use as columns
            indices: The indices associated with items

        Returns:
            The heading and menu items
        """
        menu_builder = MenuBuilder(
            progress_bar_width=self._progress_bar_width,
            screen_width=self._screen_width,
            number_colors=curses.COLORS,
            color_menu_item=self._color_menu_item,
            ui_config=self._ui_config,
        )
        menu_heading, menu_items = menu_builder.build(current, columns, indices)
        return menu_heading, menu_items

    def _handle_menu_cursor_keys(
        self,
        entry: str,
        menu_heading: CursesLines,
        showing_indices: tuple[int, ...],
    ) -> None:
        """Handle KEY_DOWN/KEY_UP cursor movement in a menu.

        Updates ``_menu_cursor_pos`` and adjusts scroll so the highlighted
        row stays visible.

        Args:
            entry: The key that was pressed ("KEY_DOWN" or "KEY_UP")
            menu_heading: The current menu heading lines (used for
                viewport height calculation)
            showing_indices: The indices currently visible on screen
        """
        # Activate cursor at position 0 on first arrow-key press (no movement)
        if self._menu_cursor_pos is None:
            self._menu_cursor_pos = 0
            return
        # Move the cursor position (only if cursor was already active)
        if entry == "KEY_DOWN" and self._menu_cursor_pos < len(self._menu_indices) - 1:
            self._menu_cursor_pos += 1
            # If moved past the last visible item, scroll down one
            if (
                self._highlight_line_offset is None
                or self._highlight_line_offset >= len(showing_indices) - 1
            ):
                viewport_height = self._screen_height - len(menu_heading) - 1
                self.scroll(
                    max(
                        min(self.scroll() + 1, len(self._menu_indices)),
                        viewport_height,
                    ),
                )
        elif entry == "KEY_UP" and self._menu_cursor_pos > 0:
            self._menu_cursor_pos -= 1
            # If moved before the first visible item, scroll up one
            if self._highlight_line_offset is None or self._highlight_line_offset <= 0:
                viewport_height = self._screen_height - len(menu_heading) - 1
                self.scroll(max(self.scroll() - 1, viewport_height))

    def _show_menu(
        self,
        current: Sequence[Any],
        columns: list[str],
        await_input: bool,
    ) -> Interaction:
        """Show a menu on the screen.

        Args:
            current: A dict
            columns: The keys from the dictionary to use as columns
            await_input: Should we wait for user input?

        Returns:
            Interaction with the user
        """
        # Reset cursor only when entering a new menu, not on refresh cycles
        if id(current) != self._menu_current_id:
            self._menu_cursor_pos = None
            self._menu_current_id = id(current)
        while True:
            if self.scroll() == 0:
                last_line_idx = min(len(current) - 1, self._screen_height - 3)
            else:
                last_line_idx = min(len(current) - 1, self._scroll - 1)

            first_line_idx = max(0, last_line_idx - (self._screen_height - 3))

            if self.menu_filter():
                self._menu_indices = tuple(
                    idx for idx, mi in enumerate(current) if self._obj_match_filter(mi, columns)
                )
                line_numbers = tuple(range(last_line_idx - first_line_idx + 1))
                self._scroll = min(len(self._menu_indices), self._scroll)
            else:
                self._menu_indices = tuple(range(len(current)))
                line_numbers = self._menu_indices[first_line_idx : last_line_idx + 1]

            showing_indices = self._menu_indices[first_line_idx : last_line_idx + 1]
            menu_heading, menu_lines = self._get_heading_menu_items(
                current,
                columns,
                showing_indices,
            )

            # Determine which row to highlight (only when cursor is active)
            self._highlight_line_offset = None
            if self._menu_indices and self._menu_cursor_pos is not None:
                self._menu_cursor_pos = max(
                    0,
                    min(self._menu_cursor_pos, len(self._menu_indices) - 1),
                )
                selected_global_index = self._menu_indices[self._menu_cursor_pos]
                try:
                    self._highlight_line_offset = showing_indices.index(selected_global_index)
                except ValueError:
                    self._highlight_line_offset = None

            entry = self._display(
                lines=menu_lines,
                line_numbers=line_numbers,
                count=len(self._menu_indices),
                heading=menu_heading,
                indent_heading=True,
                key_dict={"[0-9]": "goto"},
                await_input=await_input,
            )

            # Handle arrow navigation for menus when enabled
            if entry in ["KEY_RESIZE", "KEY_DOWN", "KEY_UP", "KEY_NPAGE", "KEY_PPAGE", "^F", "^B"]:
                if entry in ["KEY_DOWN", "KEY_UP"] and self._menu_indices:
                    self._handle_menu_cursor_keys(entry, menu_heading, showing_indices)
                continue

            # Enter key selects the highlighted item, but only when cursor is active.
            # If cursor is inactive (None), ignore the Enter entirely so it does not
            # fall through to _template_match_action and trigger a spurious warning dialog.
            if entry in ["CURSOR_ENTER", "^J", "^M", "KEY_ENTER", "KEY_RETURN"]:
                if self._menu_cursor_pos is not None and self._menu_indices:
                    entry = str(self._menu_indices[self._menu_cursor_pos])
                else:
                    continue

            name, action = self._template_match_action(entry, current)
            if name and action:
                if name == "select":
                    if current:
                        index = self._menu_indices[int(entry) % len(self._menu_indices)]
                        action = action._replace(value=index)
                    else:
                        continue
                menu = Menu(current=current, columns=columns)
                return Interaction(name=name, action=action, menu=menu, ui=self._ui)

    def show(
        self,
        obj: ContentType,
        content_format: ContentFormat | None = None,
        index: int | None = None,
        columns: list[str] | None = None,
        await_input: bool = True,
        filter_content_keys: Callable[..., Any] = lambda x: x,
        color_menu_item: Callable[..., Any] = lambda *args, **kwargs: (0, 0),
        content_heading: Callable[..., Any] = lambda *args, **kwargs: None,
    ) -> Interaction:
        """Show something on the screen.

        Args:
            obj: The inbound object
            content_format: Set the content format
            index: When obj is a list, show this entry
            columns: When obj is a list of dicts, use these keys for
                menu columns
            await_input: Should we wait for user input?
            filter_content_keys: To show the filter content keys
            color_menu_item: To show the colored menu item
            content_heading: Show the content heading

        Returns:
            interaction with the user
        """
        self._color_menu_item = color_menu_item
        self._content_heading = content_heading
        self._filter_content_keys = filter_content_keys
        columns = columns or []
        self.content_format(content_format or self._default_content_format)

        if index is not None and isinstance(obj, list | tuple):
            result = self._show_obj_from_list(obj, index, await_input)
        elif columns and isinstance(obj, list | tuple):
            result = self._show_menu(obj, columns, await_input)
        else:
            result = self._show_obj_from_list([obj], 0, await_input)
        return result

    def show_form(self, form: Form) -> Form:
        """Show a form on using the user interface.

        Args:
            form: The form to show

        Returns:
            The form populated with the response
        """
        form_result = self._show_form(form)
        return form_result



"""The configuration object."""

from __future__ import annotations

import logging
import os

from copy import deepcopy
from pathlib import Path
from typing import Any

from ansible_navigator.utils.definitions import ExitMessage
from ansible_navigator.utils.definitions import ExitPrefix
from ansible_navigator.utils.definitions import LogMessage
from ansible_navigator.utils.functions import oxfordcomma
from ansible_navigator.utils.functions import shlex_join
from ansible_navigator.utils.json_schema import validate
from ansible_navigator.utils.serialize import SafeLoader
from ansible_navigator.utils.serialize import yaml
from ansible_navigator.utils.version_migration.definitions import MigrationType
from ansible_navigator.utils.version_migration.migrate import run_all_migrations

from .definitions import ApplicationConfiguration
from .definitions import Constants as C
from .definitions import SettingsEntry
from .parser import Parser
from .transform import to_schema
from .utils import parse_ansible_cfg


class Configurator:
    """The configuration class."""

    def __init__(
        self,
        params: list[str] | None,
        application_configuration: ApplicationConfiguration,
        apply_previous_cli_entries: list[str] | C = C.NONE,
        skip_roll_back: bool = False,
    ) -> None:
        """Initialize the configuration variables.

        Args:
            params: A list of parameters e.g. ['-x', 'value']
            application_configuration: An application specific Config
                object
            apply_previous_cli_entries: Apply previous USER_CLI values
                where the current value is not a USER_CLI sourced value,
                a list of entry names ['all'] will apply all previous
            skip_roll_back: Skip roll back on error
        """
        self._apply_previous_cli_entries = apply_previous_cli_entries
        self._config = application_configuration
        self._exit_messages: list[ExitMessage] = []
        self._messages: list[LogMessage] = []
        self._params = params if params is not None else []
        self._sanity_check()
        self._skip_rollback = skip_roll_back
        self._unaltered_entries = deepcopy(self._config.entries)

    def _sanity_check(self) -> None:
        """Sanity check the configuration.

        Raises:
            ValueError: If apply_previous_cli is used while initializing
        """
        if self._apply_previous_cli_entries is not C.NONE:
            if self._config.internals.initializing:
                msg = "'apply_previous_cli' cannot be used while initializing"
                raise ValueError(msg)
            if not self._config.initial:
                msg = "'apply_previous_cli' enabled prior to an initialization"
                raise ValueError(msg)

    def _roll_back(self) -> None:
        """In the case of a rollback, log the configuration state prior to roll back."""
        if self._skip_rollback:
            return
        message = "Configuration errors encountered, rolling back to previous configuration."
        self._messages.append(LogMessage(level=logging.WARNING, message=message))
        for entry in self._config.entries:
            message = f"Prior to rollback: {entry.name} = '{entry.value.current}'"
            message += f" ({type(entry.value.current).__name__}/{entry.value.source.value})"
            self._messages.append(LogMessage(level=logging.DEBUG, message=message))
        self._config.entries = self._unaltered_entries
        for entry in self._config.entries:
            message = f"After rollback: {entry.name} = '{entry.value.current}'"
            message += f" ({type(entry.value.current).__name__}/{entry.value.source.value})"
            self._messages.append(LogMessage(level=logging.DEBUG, message=message))
        message = "Configuration rollback complete."
        self._messages.append(LogMessage(level=logging.DEBUG, message=message))

    def configure(self) -> tuple[list[LogMessage], list[ExitMessage]]:
        """Perform the configuration.

        Save the original entries, if an error is encountered
        restore them.

        Returns:
            Log messages
        """
        self._config.original_command = self._params
        shlex_joined = shlex_join(self._config.original_command)
        cmd_message = f"Command provided: '{shlex_joined}'"
        self._messages.append(LogMessage(level=logging.DEBUG, message=cmd_message))
        warn_message = "Issues were found while applying the settings."
        warning = ExitMessage(message=warn_message, prefix=ExitPrefix.WARNING)
        command = ExitMessage(message=cmd_message, prefix=ExitPrefix.HINT)

        self._restore_original()
        self._apply_defaults()
        self._apply_settings_file()
        self._apply_environment_variables()
        self._apply_cli_params()
        if self._exit_messages:
            self._exit_messages[0:0] = [warning, command]
            self._roll_back()
            return self._messages, self._exit_messages

        self._apply_previous_cli_to_current()

        self._retrieve_ansible_cfg()
        self._post_process()
        self._check_choices()
        if self._exit_messages:
            self._exit_messages[0:0] = [warning, command]
            self._roll_back()
            return self._messages, self._exit_messages

        if self._config.internals.initializing:
            self._config.initial = deepcopy(self._config)
            # Our work is done, set the initialization flag to false
            self._config.internals.initializing = False

        return self._messages, self._exit_messages

    def _argparse_error_handler(self, message: str) -> None:
        """Call back for argparser error handling.

        Args:
            message (str): A message from the parser
        """
        self._exit_messages.append(ExitMessage(message=message))

    def _restore_original(self) -> None:
        """Restore the current values back to NOT_SET."""
        for entry in self._config.entries:
            if self._config.internals.initializing or entry.change_after_initial:
                entry.value.current = C.NOT_SET
                entry.value.source = C.NOT_SET
            else:
                message = f"'{entry.name}' cannot be reconfigured. (restore original)"
                self._messages.append(LogMessage(level=logging.INFO, message=message))

    def _apply_defaults(self) -> None:
        """Apply the default values."""
        for entry in self._config.entries:
            if self._config.internals.initializing or entry.change_after_initial:
                if entry.value.default is not C.NOT_SET:
                    entry.value.current = entry.value.default
                    entry.value.source = C.DEFAULT_CFG
            else:
                message = f"'{entry.name}' cannot be reconfigured. (apply defaults)"
                self._messages.append(LogMessage(level=logging.INFO, message=message))

    def _load_settings_yaml(self, settings_filesystem_path: str) -> dict[str, Any] | None:
        """Load and parse the YAML settings file.

        Args:
            settings_filesystem_path: Path to the settings file.

        Returns:
            The parsed config dict, or None if loading failed or should be skipped.
        """

        def _raise_value_error(message: str) -> None:
            raise ValueError(message)

        with Path(settings_filesystem_path).open(encoding="utf-8") as fh:
            try:
                config = yaml.load(fh, Loader=SafeLoader)
                if config is None:
                    if self._params in (["settings", "--sample"], ["settings", "--gs"]):
                        return None
                    _raise_value_error("Settings file cannot be empty.")
            except (yaml.scanner.ScannerError, yaml.parser.ParserError, ValueError) as exc:
                exit_msg = f"Settings file found {settings_filesystem_path}, but failed to load it."
                self._exit_messages.append(ExitMessage(message=exit_msg))
                exit_msg = f"  error was: '{' '.join(str(exc).splitlines())}'"
                self._exit_messages.append(ExitMessage(message=exit_msg))
                exit_msg = (
                    f"Try checking the settings file '{settings_filesystem_path}'"
                    "and ensure it is properly formatted"
                )
                self._exit_messages.append(
                    ExitMessage(message=exit_msg, prefix=ExitPrefix.HINT),
                )
                return None
        return config

    def _validate_settings_schema(
        self, settings_filesystem_path: str, config: dict[str, Any]
    ) -> bool:
        """Validate the settings config against the schema.

        Args:
            settings_filesystem_path: Path to the settings file.
            config: The parsed config dict.

        Returns:
            True if validation errors were found.
        """
        schema = to_schema(settings=self._config)
        errors = validate(schema=schema, data=config)
        if not errors:
            return False
        msg = f"The following errors were found in the settings file ({settings_filesystem_path}):"
        self._exit_messages.append(ExitMessage(message=msg))
        self._exit_messages.extend(error.to_exit_message() for error in errors)
        hint = "Check the settings file and compare it to the current version."
        self._exit_messages.append(ExitMessage(message=hint, prefix=ExitPrefix.HINT))
        hint = (
            "The current version can be found here:"
            " (https://ansible-navigator.readthedocs.io/en/latest/settings/"
            "#ansible-navigator-settings)"
        )
        self._exit_messages.append(ExitMessage(message=hint, prefix=ExitPrefix.HINT))
        hint = (
            "The schema used for validation can be seen with 'ansible-navigator settings --schema'"
        )
        self._exit_messages.append(ExitMessage(message=hint, prefix=ExitPrefix.HINT))
        hint = "A sample settings file can be created with 'ansible-navigator settings --sample'"
        self._exit_messages.append(ExitMessage(message=hint, prefix=ExitPrefix.HINT))
        return True

    def _apply_settings_to_entries(
        self, settings_filesystem_path: str, config: dict[str, Any]
    ) -> None:
        """Apply parsed settings config values to the configuration entries.

        Args:
            settings_filesystem_path: Path to the settings file.
            config: The parsed config dict.
        """
        for entry in self._config.entries:
            settings_file_path = entry.settings_file_path(self._config.application_name)
            path_parts = settings_file_path.split(".")
            data = config
            try:
                for key in path_parts:
                    data = data[key]
                if self._config.internals.initializing or entry.change_after_initial:
                    entry.value.current = data
                    entry.value.source = C.USER_CFG
                else:
                    message = f"'{entry.name}' cannot be reconfigured. (settings file)"
                    self._messages.append(LogMessage(level=logging.INFO, message=message))
            except TypeError as exc:
                exit_msg = (
                    "Errors encountered when loading settings file:"
                    f" {settings_filesystem_path}"
                    f" while loading entry {entry.name}, attempted: {settings_file_path}."
                    f"The resulting error was {exc!s}"
                )
                self._exit_messages.append(ExitMessage(message=exit_msg))
                exit_msg = (
                    f"Try checking the settings file '{settings_filesystem_path}'"
                    "and ensure it is properly formatted"
                )
                self._exit_messages.append(
                    ExitMessage(message=exit_msg, prefix=ExitPrefix.HINT),
                )
                return
            except KeyError:
                message = f"{settings_file_path} not found in settings file"
                self._messages.append(LogMessage(level=logging.DEBUG, message=message))

    def _apply_settings_file(self) -> None:
        """Apply the settings file."""
        settings_filesystem_path = self._config.internals.settings_file_path

        if not isinstance(settings_filesystem_path, str):
            return

        run_all_migrations(
            settings_file_str=settings_filesystem_path,
            migration_types=(MigrationType.SETTINGS_FILE,),
        )

        config = self._load_settings_yaml(settings_filesystem_path)
        if config is None:
            return

        if self._validate_settings_schema(settings_filesystem_path, config):
            return

        self._apply_settings_to_entries(settings_filesystem_path, config)

    def _apply_environment_variables(self) -> None:
        """Apply the environment variables."""
        for entry in self._config.entries:
            set_env_var = os.environ.get(entry.environment_variable(self._config.application_name))
            if set_env_var is not None:
                if self._config.internals.initializing or entry.change_after_initial:
                    if entry.cli_parameters is not None and entry.cli_parameters.nargs in [
                        "+",
                        "*",
                    ]:
                        entry.value.current = [
                            value.strip()
                            for value in set_env_var.split(entry.environment_variable_split_char)
                        ]
                    else:
                        entry.value.current = set_env_var
                    entry.value.source = C.ENVIRONMENT_VARIABLE
                else:
                    message = f"'{entry.name}' cannot be reconfigured. (environment variables)"
                    self._messages.append(LogMessage(level=logging.INFO, message=message))

    def _apply_cli_params(self) -> None:
        """Apply the cli params."""
        parser = Parser(self._config).parser
        setattr(parser, "error", self._argparse_error_handler)  # noqa: B010
        try:
            # split the _params on double-dash
            index = self._params.index("--")
            before, after = self._params[:index], self._params[index + 1 :]
        except ValueError:
            before, after = self._params, []

        parser_response = parser.parse_known_args(before)
        if parser_response is None:
            return
        args, unknown = parser_response
        cmdline = unknown + after
        if cmdline:
            # In the case a subcommand is not a positional, remove the --
            additional_args = [arg for arg in cmdline if arg != "--"]
            self._config.entry("cmdline").value.current = additional_args
            self._config.entry("cmdline").value.source = C.USER_CLI
        for param, value in vars(args).items():
            if self._config.entry(param).subcommand_value is True and value is None:
                continue
            entry = self._config.entry(param)
            if self._config.internals.initializing or entry.change_after_initial:
                entry.value.current = value
                entry.value.source = C.USER_CLI
            else:
                message = f"'{entry.name}' cannot be reconfigured. (cli params)"
                self._messages.append(LogMessage(level=logging.INFO, message=message))

    def _post_process(self) -> None:
        """Post process the settings entries."""
        delayed = []
        normal = []

        # Separate normal and delayed entries so they can be processed in that order.
        for entry in self._config.entries:
            if entry.delay_post_process:
                delayed.append(entry)
            else:
                normal.append(entry)

        for entry in normal + delayed:
            if self._config.internals.initializing or entry.change_after_initial:
                processor = getattr(self._config.post_processor, entry.name, None)
                if callable(processor):
                    messages, errors = processor(entry=entry, config=self._config)
                    self._messages.extend(messages)
                    self._exit_messages.extend(errors)
            else:
                message = f"'{entry.name}' cannot be reconfigured. (post process)"
                self._messages.append(LogMessage(level=logging.INFO, message=message))

    def _check_choices(self) -> None:
        """Check the choices for each settings entry."""
        for entry in self._config.entries:
            if entry.cli_parameters and entry.choices:
                self._check_entry_choices(entry)

    def _check_entry_choices(self, entry: SettingsEntry) -> None:
        """Check choice validity for a single settings entry.

        Args:
            entry: The settings entry to check.
        """
        if isinstance(entry.value.current, list):
            for value in entry.value.current:
                logged = self._check_choice(entry=entry, value=value)
                if logged:
                    break
        else:
            self._check_choice(entry=entry, value=entry.value.current)

    def _check_choice(self, entry: SettingsEntry, value: bool | str) -> bool:
        """Check the choice for a single settings entry.

        Args:
            entry: The settings entry to check.
            value: The value to check.

        Returns:
            True if the value is invalid and an error message was
            logged.
        """
        if entry.cli_parameters and entry.choices and value not in entry.choices:
            self._exit_messages.append(ExitMessage(message=entry.invalid_choice))
            choices = [
                f"{entry.cli_parameters.short} {str(choice).lower()}" for choice in entry.choices
            ]
            exit_msg = f"Try again with {oxfordcomma(choices, 'or')}"
            self._exit_messages.append(
                ExitMessage(message=exit_msg, prefix=ExitPrefix.HINT),
            )
            return True
        return False

    def _is_eligible_for_previous_cli(
        self,
        current_entry: SettingsEntry,
        previous_entry: SettingsEntry,
        current_subcommand: str,
        previous_subcommand: str,
    ) -> bool:
        """Check if a previous CLI entry is eligible for reapplication.

        Args:
            current_entry: The current settings entry.
            previous_entry: The corresponding previous settings entry.
            current_subcommand: The current subcommand name.
            previous_subcommand: The previous subcommand name.

        Returns:
            True if the entry is eligible for reapplication.
        """
        if not any((self._config.internals.initializing, current_entry.change_after_initial)):
            message = f"'{current_entry.name}' cannot be reconfigured (apply previous cli)"
            self._messages.append(LogMessage(level=logging.INFO, message=message))
            return False
        if current_entry.value.source is C.USER_CLI:
            return False
        if (
            isinstance(self._apply_previous_cli_entries, list)
            and current_entry.name not in self._apply_previous_cli_entries
        ):
            return False
        if previous_entry.apply_to_subsequent_cli not in [C.ALL, C.SAME_SUBCOMMAND]:
            return False
        if (
            current_entry.apply_to_subsequent_cli is C.SAME_SUBCOMMAND
            and current_subcommand != previous_subcommand
        ):
            return False
        return previous_entry.value.source is C.USER_CLI

    def _apply_previous_cli_to_current(self) -> None:
        """Apply eligible previous CLI values to current not set by the CLI."""
        if self._apply_previous_cli_entries is not C.ALL and not isinstance(
            self._apply_previous_cli_entries,
            list,
        ):
            return

        current_subcommand = next(
            entry.value.current for entry in self._config.entries if entry.subcommand_value is True
        )
        previous_subcommand = next(
            entry.value.current
            for entry in self._config.initial.entries
            if entry.subcommand_value is True
        )

        for current_entry in self._config.entries:
            previous_entry = self._config.initial.entry(current_entry.name)
            if self._is_eligible_for_previous_cli(
                current_entry,
                previous_entry,
                current_subcommand,
                previous_subcommand,
            ):
                current_entry.value.current = previous_entry.value.current
                current_entry.value.source = C.PREVIOUS_CLI

    def _retrieve_ansible_cfg(self) -> None:
        """Retrieve the ansible.cfg file.

        EE support is needed early on here so the post processors
        can have access to the ansible.cfg file contents as a fallback to
        navigators settings sources. The value won't be set but it is needed to
        determine where the ansible.cfg file should be pulled from
        """
        ee_enabled = str(self._config.execution_environment).lower() == "true"
        parsed_ansible_cfg = parse_ansible_cfg(ee_enabled=ee_enabled)
        self._messages.extend(parsed_ansible_cfg.messages)
        self._exit_messages.extend(parsed_ansible_cfg.exit_messages)
        self._config.internals.ansible_configuration = parsed_ansible_cfg.config



"""Simple utils for working with forms."""

from __future__ import annotations

import copy
import shutil
import textwrap

from functools import partial
from typing import Any

from ansible_navigator.utils.definitions import ExitMessage
from ansible_navigator.utils.definitions import ExitMessages
from ansible_navigator.utils.definitions import ExitPrefix
from ansible_navigator.utils.functions import console_width

from .colorize import ansi_to_curses
from .curses_defs import CursesLines
from .field_checks import FieldChecks
from .field_curses_information import FieldCursesInformation
from .field_information import FieldInformation
from .field_option import FieldOption
from .field_radio import FieldRadio
from .field_text import FieldText
from .field_working import FieldWorking
from .form import Form
from .form import FormType
from .ui_constants import Color
from .validators import FieldValidators


def _build_text_field(field: dict[str, Any]) -> FieldText:
    """Build a FieldText from a field dictionary.

    Args:
        field: Field dictionary from form data

    Returns:
        Constructed FieldText instance
    """
    field_params: dict[str, Any] = {"name": field["name"]}
    field_params["prompt"] = field["prompt"]
    field_params["validator"] = getattr(FieldValidators, field["validator"]["name"])

    choices = field["validator"].get("choices")
    if choices:
        field_params["validator"] = partial(field_params["validator"], choices=choices)

    default = field.get("default")
    if default:
        field_params["default"] = default

    frm_field_text = FieldText(**field_params)

    pre_populate = field.get("pre_populate")
    if pre_populate:
        frm_field_text.pre_populate(pre_populate)

    return frm_field_text


def _build_checks_field(field: dict[str, Any]) -> FieldChecks:
    """Build a FieldChecks from a field dictionary.

    Args:
        field: Field dictionary from form data

    Returns:
        Constructed FieldChecks instance
    """
    field_params: dict[str, Any] = {"name": field["name"]}
    field_params["prompt"] = field["prompt"]
    field_params["options"] = [FieldOption(**option) for option in field["options"]]

    max_selected = field.get("max_selected")
    if max_selected:
        field_params["max_selected"] = max_selected

    min_selected = field.get("min_selected")
    if min_selected:
        field_params["min_selected"] = min_selected

    return FieldChecks(**field_params)


def _build_radio_field(field: dict[str, Any]) -> FieldRadio:
    """Build a FieldRadio from a field dictionary.

    Args:
        field: Field dictionary from form data

    Returns:
        Constructed FieldRadio instance
    """
    field_params: dict[str, Any] = {"name": field["name"]}
    field_params["prompt"] = field["prompt"]
    field_params["options"] = [FieldOption(**option) for option in field["options"]]
    return FieldRadio(**field_params)


def dict_to_form(form_data: dict[str, Any]) -> Form:
    """Convert a python dict to a form.

    Args:
        form_data: Form data

    Returns:
        Object containing fields from form
    """
    if form_data.get("type") == "notification":
        form = Form(type_=FormType.NOTIFICATION)
    elif form_data.get("type") == "working":
        form = Form(type_=FormType.WORKING)
    else:
        form = Form(type_=FormType.FORM)

    form._dict = form_data  # noqa: SLF001
    form.title = form_data["title"]
    form.title_color = form_data.get("title_color", 0)

    for field in form_data["fields"]:
        if field["type"] == "text_input":
            form.fields.append(_build_text_field(field))

        elif field["type"] == "checkbox":
            form.fields.append(_build_checks_field(field))

        elif field["type"] == "radio":
            form.fields.append(_build_radio_field(field))

        elif field["type"] == "information":
            frm_field_info = FieldInformation(name=field["name"], information=field["information"])
            form.fields.append(frm_field_info)

        elif field["type"] == "working":
            frm_field_working = FieldWorking(name=field["name"], messages=field["messages"])
            form.fields.append(frm_field_working)

    return form


def form_to_dict(form: Form, key_on_name: bool = False) -> dict[str, Any]:
    """Populate the original _dict of the form with the results.

    Args:
        form: Holding place for form fields
        key_on_name: Bool used to filter via name

    Returns:
        form as type dict
    """
    res = form._dict  # noqa: SLF001
    res["cancelled"] = form.cancelled
    res["submitted"] = form.submitted
    for field_idx, field in enumerate(form.fields):
        if isinstance(field, FieldText):
            res["fields"][field_idx]["response"] = copy.copy(field.response)
            res["fields"][field_idx]["value"] = copy.copy(field.value)
        elif isinstance(field, FieldChecks | FieldRadio):
            for option_idx, option in enumerate(field.options):
                res_field_options = res["fields"][field_idx]["options"]
                res_field_options[option_idx]["checked"] = option.checked
            res["fields"][field_idx]["checked"] = [
                option.name for option in field.options if option.checked
            ]

    if key_on_name:
        res["fields"] = _rekey_fields_by_name(res)
    return res


def _rekey_fields_by_name(res: dict[str, Any]) -> dict[str, Any]:
    """Rekey the fields list into a dict keyed by field name.

    Args:
        res: The form result dictionary containing a "fields" list

    Returns:
        Dictionary of fields keyed by name
    """
    fields: dict[str, Any] = {}
    for field in res["fields"]:
        fields[field["name"]] = field
        if "options" in field:
            options = {}
            for option in field["options"]:
                options[option["name"]] = option
            fields[field["name"]]["options"] = options
    return fields


def break_long_lines(messages: list[str]) -> list[str]:
    """Break lines such that the form width !> 80%.

    Args:
        messages: Lines of the form

    Returns:
        Resulting messages split if needed
    """
    width = int(shutil.get_terminal_size().columns * 0.8)
    result = []
    for message in messages:
        lns = textwrap.wrap(message, width)
        result.extend(lns)
    return result


def nonblocking_notification(messages: list[str]) -> Form:
    """Generate a std nonblocking notification.

    Args:
        messages: List of messages to display

    Returns:
        Form to display as type dict
    """
    messages = break_long_lines(messages)
    form = {
        "title": "Working on it...",
        "title_color": 2,
        "fields": [{"name": "message", "messages": messages, "type": "working"}],
        "type": "working",
    }
    return dict_to_form(form)


def settings_notification(color: bool, messages: list[ExitMessage]) -> Form:
    """Generate a warning notification for settings errors.

    Args:
        messages: List of messages to display
        color: Bool to reflect if color is transferred or not

    Returns:
        The form to display
    """
    # Take the initial warning if there is one
    if messages[0].prefix is ExitPrefix.WARNING:
        title = messages.pop(0).to_lines(color=False, width=console_width(), with_prefix=True)[0]
    else:
        title = "Warning"

    formatted = ExitMessages(messages).to_strings(color=color, width=console_width())
    formatted_curses = CursesLines(
        tuple(ansi_to_curses(line) for line in formatted),
    )
    form = Form(
        FormType.NOTIFICATION,
        title=title,
        title_color=Color.YELLOW,
        fields=[
            FieldCursesInformation(
                name="settings_warning",
                information=formatted_curses,
            ),
        ],
    )
    return form


def warning_notification(messages: list[str]) -> Form:
    """Generate a std warning notification.

    Args:
        messages: List of warning messages to be displayed

    Returns:
        Form to display as type dict
    """
    messages = break_long_lines(messages)
    form = {
        "title": "WARNING",
        "title_color": 3,
        "fields": [{"name": "info", "information": messages, "type": "information"}],
        "type": "notification",
    }
    return dict_to_form(form)


def error_notification(messages: list[str]) -> Form:
    """Generate a std error notification.

    Args:
        messages: List of error messages to display

    Returns:
        Form to display as type dict
    """
    messages = break_long_lines(messages)
    form = {
        "title": "ERROR",
        "title_color": 1,
        "fields": [{"name": "info", "information": messages, "type": "information"}],
        "type": "notification",
    }
    return dict_to_form(form)


def success_notification(messages: list[str]) -> Form:
    """Generate a std success notification.

    Args:
        messages: List of success messages to display

    Returns:
        Form to display as type dict
    """
    messages = break_long_lines(messages)
    form = {
        "title": "SUCCESS",
        "title_color": 2,
        "fields": [{"name": "info", "information": messages, "type": "information"}],
        "type": "notification",
    }
    return dict_to_form(form)



"""``:help`` command implementation."""

from ansible_navigator.action_base import ActionBase
from ansible_navigator.app_public import AppPublic
from ansible_navigator.configuration_subsystem.definitions import ApplicationConfiguration
from ansible_navigator.content_defs import ContentFormat
from ansible_navigator.ui_framework import Interaction
from ansible_navigator.utils.packaged_data import retrieve_content

from . import _actions as actions


@actions.register
class Action(ActionBase):
    """``:help`` command implementation."""

    KEGEX = r"^h(?:elp)?$"

    def __init__(self, args: ApplicationConfiguration) -> None:
        """Initialize the ``:help`` action.

        Args:
            args: The current settings for the application
        """
        super().__init__(args=args, logger_name=__name__, name="help")

    def run(self, interaction: Interaction, app: AppPublic) -> Interaction:
        """Execute the ``:help`` request.

        Args:
            interaction: The interaction from the user
            app: The app instance

        Returns:
            The pending
            :class:`~ansible_navigator.ui_framework.ui.Interaction`
        """
        self._logger.debug("help requested")
        self._prepare_to_run(app, interaction)

        help_md = retrieve_content(filename="help.md")
        while True:
            interaction = interaction.ui.show(
                obj=help_md,
                content_format=ContentFormat.MARKDOWN,
            )
            app.update()
            if interaction.name != "refresh":
                break

        self._prepare_to_exit(interaction)
        return interaction



"""``:filter`` command implementation."""

import logging

from ansible_navigator.app_public import AppPublic
from ansible_navigator.configuration_subsystem.definitions import ApplicationConfiguration
from ansible_navigator.ui_framework import Interaction

from . import _actions as actions


@actions.register
class Action:
    """``:filter`` command implementation."""

    KEGEX = r"^f(ilter)?(\s(?P<regex>.*))?$"

    def __init__(self, args: ApplicationConfiguration) -> None:
        """Initialize the ``:filter`` action.

        Args:
            args: The current settings for the application
        """
        self._args = args
        self._logger = logging.getLogger(__name__)

    def run(self, interaction: Interaction, app: AppPublic) -> None:
        """Execute the ``:filter`` request.

        Args:
            interaction: The interaction from the user
            app: The app instance
        """
        self._logger.debug("filter requested")
        menu_filter = interaction.action.match.groupdict()["regex"]
        interaction.ui.menu_filter(menu_filter)
        self._logger.debug("requested filter set to %s", menu_filter)



"""Stub (interface) file for ``_version.py``.

Provides type hint PEP 484 information for the ``_version.py``
file because it is generated by setuptools_scm and may
not exist when type checking is done and will not
contain type hint information.
"""  # noqa: PYI021

version: str



"""Run the :exec subcommand."""

from __future__ import annotations

import logging
import shlex

from pathlib import Path

from ansible_navigator.action_base import ActionBase
from ansible_navigator.action_defs import RunStdoutReturn
from ansible_navigator.configuration_subsystem.definitions import ApplicationConfiguration
from ansible_navigator.configuration_subsystem.definitions import Constants
from ansible_navigator.runner import Command

from . import _actions as actions


GeneratedCommand = tuple[str, list[str] | None]

logger = logging.getLogger(__name__)


def _generate_command(
    exec_command: str,
    exec_shell: bool,
    extra_args: Constants | list[str],
    exec_command_is_default: bool,
) -> GeneratedCommand:
    """Generate the command and args.

    Args:
        exec_command: The command to run
        exec_shell: Should the command be wrapped in a shell
        extra_args: Any unknown or extra arguments passed on the command
            line
        exec_command_is_default: If the exec_command is set to its
            default value

    Returns:
        The command and any pass through arguments
    """
    logger.debug("exec_command: %s", exec_command)
    logger.debug("exec_shell: %s", exec_shell)
    logger.debug("extra_args: %s", extra_args)
    if exec_shell:
        command = "/bin/bash"
        # Determine if any extra args were picked up
        _extra_args = []
        if not exec_command_is_default:
            _extra_args.append(exec_command)
        if isinstance(extra_args, list):
            _extra_args.extend(extra_args)
        if _extra_args:
            pass_command = " ".join(_extra_args)
            pass_through_args = ["-c", pass_command]
        else:
            pass_through_args = []

    elif exec_command_is_default and isinstance(extra_args, list):
        command, pass_through_args = extra_args[0], extra_args[1:]
    else:
        parts = shlex.split(exec_command)
        command = parts[0]
        if len(parts) == 1 and isinstance(extra_args, list):
            # Use the extra arguments
            pass_through_args = extra_args
        else:
            # Use the leftovers or an empty list
            pass_through_args = parts[1:]
    logger.debug("runner command: %s", command)
    logger.debug("runner passthrough: %s", pass_through_args)
    return (command, pass_through_args)


@actions.register
class Action(ActionBase):
    """Run the :exec subcommand."""

    KEGEX = "^e(?:xec)?$"

    def __init__(self, args: ApplicationConfiguration) -> None:
        """Initialize the ``:exec`` action.

        Args:
            args: The current settings for the application
        """
        super().__init__(args=args, logger_name=__name__, name="exec")

    def run_stdout(self) -> RunStdoutReturn:
        """Execute the ``exec`` request for mode stdout.

        Returns:
            The return code or 1. If the response from the runner
            invocation is None, indicates there is no console output to
            display, so assume an issue and return 1 along with a
            message to review the logs.
        """
        self._logger.debug("exec requested in stdout mode")
        response = self._run_runner()
        if response is None:
            self._logger.error("Unexpected response: %s", response)
            return RunStdoutReturn(message="Please review the log for errors.", return_code=1)
        _out, error, return_code = response
        return RunStdoutReturn(message=error, return_code=return_code)

    def _run_runner(self) -> tuple[str, str, int] | None:
        """Spin up runner.

        Returns:
            The stdout, stderr and return code from runner
        """
        if isinstance(self._args.set_environment_variable, dict):
            env_vars_to_set = self._args.set_environment_variable.copy()
        elif isinstance(self._args.set_environment_variable, Constants):
            env_vars_to_set = {}
        else:
            log_message = (
                "The setting 'set_environment_variable' was neither a dictionary"
                " or Constants, please raise an issue. No environment variables will be set."
            )
            self._logger.error(
                "%s The current value was found to be '%s'",
                log_message,
                self._args.set_environment_variable,
            )
            env_vars_to_set = {}

        if self._args.display_color is False:
            env_vars_to_set["ANSIBLE_NOCOLOR"] = "1"

        kwargs = {
            "container_engine": self._args.container_engine,
            "host_cwd": str(Path.cwd()),
            "execution_environment_image": self._args.execution_environment_image,
            "execution_environment": self._args.execution_environment,
            "navigator_mode": self._args.mode,
            "pass_environment_variable": self._args.pass_environment_variable,
            "set_environment_variable": env_vars_to_set,
            "timeout": self._args.ansible_runner_timeout,
        }

        if isinstance(self._args.execution_environment_volume_mounts, list):
            kwargs["container_volume_mounts"] = self._args.execution_environment_volume_mounts

        if isinstance(self._args.container_options, list):
            kwargs["container_options"] = self._args.container_options

        command, pass_through_args = _generate_command(
            exec_command=self._args.exec_command,
            exec_shell=self._args.exec_shell,
            extra_args=self._args.cmdline,
            exec_command_is_default=self._args.entry("exec_command").value.source
            is Constants.DEFAULT_CFG,
        )
        if isinstance(pass_through_args, list):
            kwargs["cmdline"] = pass_through_args

        runner = Command(executable_cmd=command, **kwargs)
        return runner.run()



"""The ansible-navigator configuration."""

from __future__ import annotations

import logging
import os

from dataclasses import dataclass
from dataclasses import field
from pathlib import Path
from typing import TYPE_CHECKING

from ansible_navigator.utils.definitions import ExitMessage
from ansible_navigator.utils.definitions import LogMessage
from ansible_navigator.utils.functions import expand_path
from ansible_navigator.utils.functions import generate_cache_path
from ansible_navigator.utils.functions import oxfordcomma
from ansible_navigator.utils.packaged_data import ImageEntry

from .definitions import ApplicationConfiguration
from .definitions import CliParameters
from .definitions import Constants as C
from .definitions import SettingsEntry
from .definitions import SettingsEntryValue
from .definitions import SubCommand
from .navigator_post_processor import NavigatorPostProcessor
from .utils import AnsibleConfiguration


if TYPE_CHECKING:
    from ansible_navigator.utils.key_value_store import KeyValueStore


APP_NAME = "ansible_navigator"

initialization_messages: list[LogMessage] = []
initialization_exit_messages: list[ExitMessage] = []

PLUGIN_TYPES = (
    "become",
    "cache",
    "callback",
    "cliconf",
    "connection",
    "filter",
    "httpapi",
    "inventory",
    "keyword",
    "lookup",
    "module",
    "netconf",
    "role",
    "shell",
    "strategy",
    "test",
    "vars",
)


def generate_editor_command() -> str:
    """Generate a default for editor_command if EDITOR is set.

    Returns:
        command to be run by specific editor
    """
    editor = os.environ.get("EDITOR")
    if editor is None:
        message = "EDITOR environment variable not set"
        initialization_messages.append(LogMessage(level=logging.DEBUG, message=message))
        command = "vi +{line_number} {filename}"
    else:
        message = f"EDITOR environment variable set as '{editor}'"
        initialization_messages.append(LogMessage(level=logging.DEBUG, message=message))
        command = f"{editor!s} {{filename}}"
    message = f"Default editor_command set to: {command}"
    initialization_messages.append(LogMessage(level=logging.DEBUG, message=message))
    return command


def _generate_doc_cache_path() -> str:
    """Generate a path for the collection cache.

    Returns:
        Collection cache path
    """
    file_name = "collection_doc_cache.db"
    cache_dir = generate_cache_path(APP_NAME.replace("_", "-"))
    cache_path = cache_dir / file_name
    message = f"Default collection_doc_cache_path set to: {cache_path}"
    initialization_messages.append(LogMessage(level=logging.DEBUG, message=message))
    return str(cache_path)


@dataclass
class Internals:
    """Place to hold an object that needs to be used from app initiation through whole app."""

    ansible_configuration: AnsibleConfiguration = field(default_factory=AnsibleConfiguration)
    action_packages: tuple[str] = ("ansible_navigator.actions",)
    cache_path: Path = generate_cache_path(APP_NAME.replace("_", "-"))  # noqa: RUF009
    collection_doc_cache: C | KeyValueStore = C.NOT_SET
    initializing: bool = False
    initialization_exit_messages = initialization_exit_messages
    initialization_messages = initialization_messages
    settings_file_path: str | None = None
    settings_source: C = C.NOT_SET


navigator_subcommands = [
    SubCommand(
        name="builder",
        description=(
            "Build [execution environment]"
            "(https://ansible.readthedocs.io/en/latest/getting_started_ee/index.html) "
            "(container image)"
        ),
        epilog=(
            "Note: 'ansible-navigator builder' additionally supports"
            " the same parameters as the 'ansible-builder' command."
            " For more information about these, try "
            " 'ansible-navigator builder --help-builder --mode stdout'"
        ),
        version_added="v2.0",
    ),
    SubCommand(
        name="collections",
        description="Explore available collections",
        version_added="v1.0",
    ),
    SubCommand(
        name="config",
        description="Explore the current ansible configuration",
        epilog=(
            "Note: With '--mode stdout', 'ansible-navigator config' additionally supports"
            " the same parameters as the 'ansible-config' command."
            " For more information about these, try "
            " 'ansible-navigator config --help-config --mode stdout'"
        ),
        version_added="v1.0",
    ),
    SubCommand(
        name="doc",
        description="Review documentation for a module or plugin",
        epilog=(
            "Note: With '--mode stdout', 'ansible-navigator doc' additionally supports"
            " the same parameters as the 'ansible-doc' command."
            " For more information about these, try "
            " 'ansible-navigator doc --help-doc --mode stdout'"
        ),
        version_added="v1.0",
    ),
    SubCommand(
        name="exec",
        description="Run a command within an execution environment",
        epilog=(
            "Note: During development, it may become necessary to interact"
            " directly with the execution environment to review and confirm"
            " its build and behavior. All navigator settings will be applied"
            " when starting the execution environment."
        ),
        version_added="v2.0",
    ),
    SubCommand(
        name="images",
        description="Explore execution environment images",
        version_added="v1.0",
    ),
    SubCommand(
        name="inventory",
        description="Explore an inventory",
        epilog=(
            "Note: With '--mode stdout', 'ansible-navigator inventory' additionally supports"
            " the same parameters as the 'ansible-inventory' command."
            " For more information about these, try "
            " 'ansible-navigator inventory --help-inventory --mode stdout'"
        ),
        version_added="v1.0",
    ),
    SubCommand(
        name="lint",
        description="Lint a file or directory for common errors and issues",
        epilog=(
            "Note: Defaults to the current working directory."
            " If using an execution environment, ansible-lint must be installed"
            " in it. If not using an execution environment, ansible-lint must"
            " be installed on your system."
        ),
        version_added="v2.0",
    ),
    SubCommand(
        name="replay",
        description="Explore a previous run using a playbook artifact",
        version_added="v1.0",
    ),
    SubCommand(
        name="run",
        description="Run a playbook",
        epilog=(
            "Note: 'ansible-navigator run' additionally supports"
            " the same parameters as the 'ansible-playbook' command."
            " For more information about these, try "
            " 'ansible-navigator run --help-playbook --mode stdout'"
        ),
        version_added="v1.0",
    ),
    SubCommand(
        name="settings",
        description="Review the current ansible-navigator settings",
        version_added="v2.0",
    ),
    SubCommand(
        name="welcome",
        description="Start at the welcome page",
        version_added="v1.0",
    ),
]

NavigatorConfiguration = ApplicationConfiguration(
    application_name=APP_NAME,
    application_version=C.NOT_SET,
    internals=Internals(),
    post_processor=NavigatorPostProcessor(),
    subcommands=navigator_subcommands,
    entries=[
        SettingsEntry(
            name="ansible_runner_artifact_dir",
            cli_parameters=CliParameters(short="--rad"),
            settings_file_path_override="ansible-runner.artifact-dir",
            short_description="The directory path to store artifacts generated by ansible-runner",
            value=SettingsEntryValue(),
            version_added="v1.0",
        ),
        SettingsEntry(
            name="ansible_runner_rotate_artifacts_count",
            cli_parameters=CliParameters(short="--rac"),
            settings_file_path_override="ansible-runner.rotate-artifacts-count",
            short_description="Keep ansible-runner artifact directories, "
            "for last n runs, if set to 0 artifact directories "
            "won't be deleted",
            value=SettingsEntryValue(),
            version_added="v1.0",
        ),
        SettingsEntry(
            name="ansible_runner_timeout",
            cli_parameters=CliParameters(short="--rt"),
            settings_file_path_override="ansible-runner.timeout",
            short_description=(
                "The timeout value after which ansible-runner will forcefully stop the execution"
            ),
            value=SettingsEntryValue(),
            version_added="v1.0",
        ),
        SettingsEntry(
            name="ansible_runner_write_job_events",
            choices=[True, False],
            cli_parameters=CliParameters(short="--rwje", action="store_true"),
            settings_file_path_override="ansible-runner.job-events",
            short_description="Write ansible-runner job_events in the artifact directory",
            value=SettingsEntryValue(default=False),
            version_added="v2.3",
        ),
        SettingsEntry(
            name="app",
            apply_to_subsequent_cli=C.NONE,
            choices=[subcommand.name for subcommand in navigator_subcommands],
            short_description="Subcommands",
            subcommand_value=True,
            value=SettingsEntryValue(default="welcome"),
            version_added="v1.0",
        ),
        SettingsEntry(
            name="cmdline",
            apply_to_subsequent_cli=C.SAME_SUBCOMMAND,
            settings_file_path_override="ansible.cmdline",
            short_description="Extra parameters passed to the underlying ansible command"
            " (e.g. ansible-playbook, ansible-doc, etc)",
            value=SettingsEntryValue(),
            version_added="v1.0",
        ),
        SettingsEntry(
            name="collection_doc_cache_path",
            cli_parameters=CliParameters(short="--cdcp"),
            short_description="The path to collection doc cache",
            value=SettingsEntryValue(
                default=_generate_doc_cache_path(),
                schema_default="~/.cache/ansible-navigator/collection_doc_cache.db",
            ),
            version_added="v1.0",
        ),
        SettingsEntry(
            name="config",
            cli_parameters=CliParameters(short="-c"),
            environment_variable_override="ansible_config",
            settings_file_path_override="ansible.config.path",
            short_description="Specify the path to the ansible configuration file",
            subcommands=["config"],
            value=SettingsEntryValue(),
            version_added="v1.0",
        ),
        SettingsEntry(
            name="container_engine",
            choices=["auto", "podman", "docker"],
            cli_parameters=CliParameters(short="--ce"),
            settings_file_path_override="execution-environment.container-engine",
            short_description="Specify the container engine (auto=podman then docker)",
            value=SettingsEntryValue(default="auto"),
            version_added="v1.0",
        ),
        SettingsEntry(
            name="container_options",
            cli_parameters=CliParameters(action="append", nargs="+", short="--co"),
            environment_variable_override="ansible_navigator_container_options",
            settings_file_path_override="execution-environment.container-options",
            short_description="Extra parameters passed to the container engine command",
            value=SettingsEntryValue(),
            version_added="v2.0",
        ),
        SettingsEntry(
            name="display_color",
            change_after_initial=False,
            choices=[True, False],
            cli_parameters=CliParameters(short="--dc"),
            environment_variable_override="no_color",
            settings_file_path_override="color.enable",
            short_description="Enable the use of color for mode interactive and stdout",
            value=SettingsEntryValue(default=True),
            version_added="v1.0",
        ),
        SettingsEntry(
            name="editor_command",
            cli_parameters=CliParameters(short="--ecmd"),
            settings_file_path_override="editor.command",
            short_description="Specify the editor command",
            value=SettingsEntryValue(
                default=generate_editor_command(),
                schema_default="vi +{line_number} {filename}",
            ),
            version_added="v1.0",
        ),
        SettingsEntry(
            name="editor_console",
            choices=[True, False],
            cli_parameters=CliParameters(short="--econ"),
            settings_file_path_override="editor.console",
            short_description="Specify if the editor is console based",
            value=SettingsEntryValue(default=True),
            version_added="v1.0",
        ),
        SettingsEntry(
            name="enable_prompts",
            choices=[True, False],
            cli_parameters=CliParameters(short="--ep", action="store_true"),
            short_description="Enable prompts for password and in playbooks. This will set"
            " mode to stdout and disable playbook artifact creation",
            subcommands=["run"],
            value=SettingsEntryValue(default=False),
            version_added="v2.3",
        ),
        SettingsEntry(
            name="exec_command",
            cli_parameters=CliParameters(positional=True),
            settings_file_path_override="exec.command",
            short_description="Specify the command to run within the execution environment",
            subcommands=["exec"],
            value=SettingsEntryValue(default="/bin/bash"),
            version_added="v2.0",
        ),
        SettingsEntry(
            name="exec_shell",
            choices=[True, False],
            cli_parameters=CliParameters(short="--exshell"),
            settings_file_path_override="exec.shell",
            short_description="Specify the exec command should be run in a shell",
            subcommands=["exec"],
            value=SettingsEntryValue(default=True),
            version_added="v2.0",
        ),
        SettingsEntry(
            name="execution_environment",
            choices=[True, False],
            cli_parameters=CliParameters(short="--ee"),
            settings_file_path_override="execution-environment.enabled",
            short_description="Enable or disable the use of an execution environment",
            value=SettingsEntryValue(default=True),
            version_added="v1.0",
        ),
        SettingsEntry(
            name="execution_environment_image",
            cli_parameters=CliParameters(short="--eei"),
            settings_file_path_override="execution-environment.image",
            short_description="Specify the name of the execution environment image",
            value=SettingsEntryValue(
                default=ImageEntry.DEFAULT_EE.get(app_name=APP_NAME),
                schema_default=C.NONE,
            ),
            version_added="v1.0",
        ),
        SettingsEntry(
            name="execution_environment_volume_mounts",
            cli_parameters=CliParameters(action="append", nargs="+", short="--eev"),
            delay_post_process=True,
            environment_variable_split_char=";",
            settings_file_path_override="execution-environment.volume-mounts",
            short_description=(
                "Specify volume to be bind mounted within an execution environment"
                " (--eev /home/user/test:/home/user/test:Z)"
            ),
            value=SettingsEntryValue(),
            version_added="v1.0",
        ),
        SettingsEntry(
            name="format",
            choices=["json", "yaml"],
            cli_parameters=CliParameters(short="--fmt"),
            short_description="Specify the format for stdout output.",
            subcommands=["collections", "images"],
            value=SettingsEntryValue(default="yaml"),
            version_added="v2.3",
        ),
        SettingsEntry(
            name="help_builder",
            choices=[True, False],
            cli_parameters=CliParameters(short="--hb", action="store_true"),
            settings_file_path_override="ansible-builder.help",
            short_description="Help options for ansible-builder command in stdout mode",
            subcommands=["builder"],
            value=SettingsEntryValue(default=False),
            version_added="v2.0",
        ),
        SettingsEntry(
            name="help_config",
            choices=[True, False],
            cli_parameters=CliParameters(short="--hc", action="store_true"),
            settings_file_path_override="ansible.config.help",
            short_description="Help options for ansible-config command in stdout mode",
            subcommands=["config"],
            value=SettingsEntryValue(default=False),
            version_added="v1.0",
        ),
        SettingsEntry(
            name="help_doc",
            choices=[True, False],
            cli_parameters=CliParameters(short="--hd", action="store_true"),
            settings_file_path_override="ansible.doc.help",
            short_description="Help options for ansible-doc command in stdout mode",
            subcommands=["doc"],
            value=SettingsEntryValue(default=False),
            version_added="v1.0",
        ),
        SettingsEntry(
            name="help_inventory",
            choices=[True, False],
            cli_parameters=CliParameters(short="--hi", action="store_true"),
            settings_file_path_override="ansible.inventory.help",
            short_description="Help options for ansible-inventory command in stdout mode",
            subcommands=["inventory"],
            value=SettingsEntryValue(default=False),
            version_added="v1.0",
        ),
        SettingsEntry(
            name="help_playbook",
            choices=[True, False],
            cli_parameters=CliParameters(short="--hp", action="store_true"),
            settings_file_path_override="ansible.playbook.help",
            short_description="Help options for ansible-playbook command in stdout mode",
            subcommands=["run"],
            value=SettingsEntryValue(default=False),
            version_added="v1.0",
        ),
        SettingsEntry(
            name="images_details",
            choices=[
                "ansible_collections",
                "ansible_version",
                "everything",
                "os_release",
                "python_packages",
                "python_version",
                "redhat_release",
                "system_packages",
            ],
            cli_parameters=CliParameters(
                action="append",
                nargs="*",
                short="-d",
                long_override="--details",
            ),
            settings_file_path_override="images.details",
            short_description=(
                "Provide detailed information about the selected execution environment image"
            ),
            subcommands=["images"],
            value=SettingsEntryValue(default=["everything"]),
            version_added="v2.0",
        ),
        SettingsEntry(
            name="inventory",
            cli_parameters=CliParameters(action="append", nargs="*", short="-i"),
            environment_variable_override="ansible_inventory",
            settings_file_path_override="ansible.inventory.entries",
            short_description="Specify an inventory file path or comma separated host list",
            subcommands=["inventory", "run"],
            value=SettingsEntryValue(),
            version_added="v1.0",
        ),
        SettingsEntry(
            name="inventory_column",
            cli_parameters=CliParameters(action="append", nargs="+", short="--ic"),
            environment_variable_override="ansible_navigator_inventory_columns",
            settings_file_path_override="inventory-columns",
            short_description="Specify a host attribute to show in the inventory view",
            subcommands=["inventory", "run"],
            value=SettingsEntryValue(),
            version_added="v1.0",
        ),
        SettingsEntry(
            name="lint_config",
            cli_parameters=CliParameters(short="--lic"),
            environment_variable_override="ansible_lint_config",
            settings_file_path_override="ansible-lint.config",
            short_description="Specify the path to the ansible-lint configuration file",
            subcommands=["lint"],
            value=SettingsEntryValue(),
            version_added="v2.0",
        ),
        SettingsEntry(
            name="lintables",
            cli_parameters=CliParameters(positional=True),
            short_description="Path to files on which to run ansible-lint",
            settings_file_path_override="ansible-lint.lintables",
            subcommands=["lint"],
            value=SettingsEntryValue(),
            version_added="v2.0",
        ),
        SettingsEntry(
            name="log_append",
            choices=[True, False],
            cli_parameters=CliParameters(short="--la"),
            short_description=(
                "Specify if log messages should be appended to an existing log file,"
                " otherwise a new log file will be created per session"
            ),
            settings_file_path_override="logging.append",
            value=SettingsEntryValue(default=True),
            version_added="v1.0",
        ),
        SettingsEntry(
            name="log_file",
            cli_parameters=CliParameters(short="--lf"),
            short_description="Specify the full path for the ansible-navigator log file",
            settings_file_path_override="logging.file",
            value=SettingsEntryValue(
                default=str(expand_path("./ansible-navigator.log")),
                schema_default="./ansible-navigator.log",
            ),
            version_added="v1.0",
        ),
        SettingsEntry(
            name="log_level",
            choices=["debug", "info", "warning", "error", "critical"],
            cli_parameters=CliParameters(short="--ll"),
            short_description="Specify the ansible-navigator log level",
            settings_file_path_override="logging.level",
            value=SettingsEntryValue(default="warning"),
            version_added="v1.0",
        ),
        SettingsEntry(
            name="mode",
            change_after_initial=False,
            delay_post_process=True,
            choices=["stdout", "interactive"],
            cli_parameters=CliParameters(short="-m"),
            short_description="Specify the user-interface mode",
            value=SettingsEntryValue(default="interactive"),
            version_added="v1.0",
        ),
        SettingsEntry(
            name="osc4",
            choices=[True, False],
            cli_parameters=CliParameters(short="--osc4"),
            settings_file_path_override="color.osc4",
            short_description="Enable or disable terminal color changing support with OSC 4",
            value=SettingsEntryValue(default=True),
            version_added="v1.0",
        ),
        SettingsEntry(
            name="pass_environment_variable",
            cli_parameters=CliParameters(action="append", nargs="+", short="--penv"),
            environment_variable_override="ansible_navigator_pass_environment_variables",
            settings_file_path_override="execution-environment.environment-variables.pass",
            short_description=(
                "Specify an existing environment variable to be passed through"
                " to and set within the execution environment (--penv MY_VAR)"
            ),
            value=SettingsEntryValue(),
            version_added="v1.0",
        ),
        SettingsEntry(
            name="playbook",
            cli_parameters=CliParameters(positional=True),
            short_description="Specify the playbook name",
            settings_file_path_override="ansible.playbook.path",
            subcommands=["run"],
            value=SettingsEntryValue(),
            version_added="v1.0",
        ),
        SettingsEntry(
            name="playbook_artifact_enable",
            choices=[True, False],
            cli_parameters=CliParameters(short="--pae"),
            delay_post_process=True,
            settings_file_path_override="playbook-artifact.enable",
            short_description="Enable or disable the creation of artifacts for"
            " completed playbooks. Note: not compatible with '--mode stdout' when playbooks"
            " require user input",
            subcommands=["run"],
            value=SettingsEntryValue(default=True),
            version_added="v1.0",
        ),
        SettingsEntry(
            name="playbook_artifact_replay",
            cli_parameters=CliParameters(positional=True),
            settings_file_path_override="playbook-artifact.replay",
            short_description="Specify the path for the playbook artifact to replay",
            subcommands=["replay"],
            value=SettingsEntryValue(),
            version_added="v1.0",
        ),
        SettingsEntry(
            name="playbook_artifact_save_as",
            cli_parameters=CliParameters(short="--pas"),
            settings_file_path_override="playbook-artifact.save-as",
            short_description=(
                "Specify the name for artifacts created from completed playbooks."
                " The following placeholders are available: {playbook_dir}, {playbook_name},"
                " {playbook_status}, and {time_stamp}"
            ),
            subcommands=["run"],
            value=SettingsEntryValue(
                default="{playbook_dir}/{playbook_name}-artifact-{time_stamp}.json",
            ),
            version_added="v1.0",
        ),
        SettingsEntry(
            name="plugin_name",
            cli_parameters=CliParameters(positional=True),
            settings_file_path_override="ansible.doc.plugin.name",
            short_description="Specify the plugin name",
            subcommands=["doc"],
            value=SettingsEntryValue(),
            version_added="v1.0",
        ),
        SettingsEntry(
            name="plugin_type",
            choices=PLUGIN_TYPES,
            cli_parameters=CliParameters(short="-t", long_override="--type"),
            settings_file_path_override="ansible.doc.plugin.type",
            short_description=f"Specify the plugin type, {oxfordcomma(PLUGIN_TYPES, 'or')}",
            subcommands=["doc"],
            value=SettingsEntryValue(default="module"),
            version_added="v1.0",
        ),
        SettingsEntry(
            name="pull_arguments",
            cli_parameters=CliParameters(action="append", nargs="+", short="--pa"),
            settings_file_path_override="execution-environment.pull.arguments",
            short_description=(
                "Specify any additional parameters that should be added to the"
                " pull command when pulling an execution environment from a container"
                " registry. e.g. --pa='--tls-verify=false'"
            ),
            value=SettingsEntryValue(),
            version_added="v2.0",
        ),
        SettingsEntry(
            name="pull_policy",
            choices=["always", "missing", "never", "tag"],
            cli_parameters=CliParameters(short="--pp"),
            settings_file_path_override="execution-environment.pull.policy",
            short_description=(
                "Specify the image pull policy"
                " always:Always pull the image,"
                " missing:Pull if not locally available,"
                " never:Never pull the image,"
                " tag:if the image tag is 'latest', always pull the image,"
                " otherwise pull if not locally available"
            ),
            value=SettingsEntryValue(default="tag"),
            version_added="v1.0",
        ),
        SettingsEntry(
            name="set_environment_variable",
            cli_parameters=CliParameters(action="append", nargs="+", short="--senv"),
            environment_variable_override="ansible_navigator_set_environment_variables",
            settings_file_path_override="execution-environment.environment-variables.set",
            short_description=(
                "Specify an environment variable and a value to be set within the"
                " execution environment (--senv MY_VAR=42)"
            ),
            value=SettingsEntryValue(),
            version_added="v1.0",
        ),
        SettingsEntry(
            name="settings_effective",
            cli_parameters=CliParameters(
                short="--se",
                long_override="--effective",
                action="store_true",
            ),
            settings_file_path_override="settings.effective",
            short_description=(
                "Show the effective settings. Defaults, CLI parameters, environment variables,"
                " and the settings file will be combined"
            ),
            subcommands=["settings"],
            value=SettingsEntryValue(default=False),
            version_added="v2.0",
        ),
        SettingsEntry(
            name="settings_sample",
            cli_parameters=CliParameters(
                short="--gs",
                long_override="--sample",
                action="store_true",
            ),
            settings_file_path_override="settings.sample",
            short_description=("Generate a sample settings file"),
            subcommands=["settings"],
            value=SettingsEntryValue(default=False),
            version_added="v2.0",
        ),
        SettingsEntry(
            name="settings_schema",
            choices=["json"],
            cli_parameters=CliParameters(
                short="--ss",
                long_override="--schema",
                const="json",
                nargs="?",
            ),
            settings_file_path_override="settings.schema",
            short_description=(
                "Generate a schema for the settings file ('json'= draft-07 JSON Schema)"
            ),
            subcommands=["settings"],
            value=SettingsEntryValue(default="json"),
            version_added="v2.0",
        ),
        SettingsEntry(
            name="settings_sources",
            cli_parameters=CliParameters(
                short="--so",
                long_override="--sources",
                action="store_true",
            ),
            settings_file_path_override="settings.sources",
            short_description=("Show the source of each current settings entry"),
            subcommands=["settings"],
            value=SettingsEntryValue(default=False),
            version_added="v2.0",
        ),
        SettingsEntry(
            name="time_zone",
            cli_parameters=CliParameters(short="--tz"),
            environment_variable_override="TZ",
            short_description=(
                "Specify the IANA time zone to use or 'local' to use the system time zone"
            ),
            value=SettingsEntryValue(default="UTC"),
            version_added="v2.0",
        ),
        SettingsEntry(
            name="workdir",
            cli_parameters=CliParameters(short="--bwd"),
            settings_file_path_override="ansible-builder.workdir",
            short_description="Specify the path that contains ansible-builder manifest files",
            subcommands=["builder"],
            value=SettingsEntryValue(
                default=str(Path.cwd()),
                schema_default=".",
            ),
            version_added="v2.0",
        ),
    ],
)



"""Inventory subcommand implementation."""

from __future__ import annotations

import json
import shlex
import shutil

from pathlib import Path
from typing import TYPE_CHECKING
from typing import Any

from ansible_navigator.action_base import ActionBase
from ansible_navigator.action_defs import RunStdoutReturn
from ansible_navigator.content_defs import ContentFormat
from ansible_navigator.runner import AnsibleInventory
from ansible_navigator.runner import Command
from ansible_navigator.steps import Step
from ansible_navigator.ui_framework import Color
from ansible_navigator.ui_framework import CursesLine
from ansible_navigator.ui_framework import CursesLinePart
from ansible_navigator.ui_framework import CursesLines
from ansible_navigator.ui_framework import Decoration
from ansible_navigator.ui_framework import Interaction
from ansible_navigator.ui_framework import warning_notification

from . import _actions as actions
from . import run_action


if TYPE_CHECKING:
    from ansible_navigator.app_public import AppPublic
    from ansible_navigator.configuration_subsystem.definitions import ApplicationConfiguration


def color_menu(colno: int, colname: str, entry: dict[str, Any]) -> tuple[int, int]:
    """Provide a color for a inventory menu entry in one column.

    Args:
        colno: The column number
        colname: The column name
        entry: The menu entry

    Returns:
        The color and decoration
    """
    if colname in ["__name", "title", "inventory_hostname"]:
        return Color.BRIGHT_GREEN, Decoration.NORMAL
    if colname == "__taxonomy":
        return Color.BRIGHT_YELLOW, Decoration.NORMAL
    if colname == "description":
        return Color.BRIGHT_BLUE, Decoration.NORMAL
    if colname == "__type":
        if entry["__type"] == "group":
            return Color.BRIGHT_YELLOW, Decoration.NORMAL
        return Color.BRIGHT_BLUE, Decoration.NORMAL
    colors = [
        Color.BRIGHT_CYAN,
        Color.BRIGHT_MAGENTA,
        Color.CYAN,
        Color.MAGENTA,
        Color.BLUE,
        Color.YELLOW,
        Color.GREEN,
    ]
    return colors[colno % len(colors)], Decoration.NORMAL


def content_heading(obj: Any, screen_w: int) -> CursesLines | None:
    """Create a heading for inventory content.

    Args:
        obj: The content going to be shown
        screen_w: The current screen width

    Returns:
        The heading
    """
    host = obj["inventory_hostname"]
    operating_system = obj.get("ansible_network_os", obj.get("ansible_platform", ""))
    string = f"[{host}] {operating_system}"
    string = string + (" " * (screen_w - len(string) + 1))
    line_part = CursesLinePart(
        column=0,
        string=string,
        color=Color.BLACK,
        decoration=Decoration.UNDERLINE,
    )
    return CursesLines((CursesLine((line_part,)),))


def filter_content_keys(obj: dict[Any, Any]) -> dict[Any, Any]:
    """Filter out some keys when showing inventory content.

    Args:
        obj: The object from which keys should be removed

    Returns:
        The object with keys removed
    """
    return {k: v for k, v in obj.items() if not k.startswith("__")}


class MenuEntry(dict[Any, Any]):
    """A menu entry."""


class Menu(list[Any]):
    """A menu."""


@actions.register
class Action(ActionBase):
    """Inventory subcommand implementation."""

    # pylint: disable=too-many-instance-attributes

    KEGEX = r"^i(?:nventory)?(\s(?P<params>.*))?$"

    def __init__(self, args: ApplicationConfiguration) -> None:
        """Initialize the ``:images`` action.

        Args:
            args: The current settings for the application
        """
        super().__init__(args=args, logger_name=__name__, name="inventory")

        self.__inventory: dict[Any, Any] = {}
        self._host_vars: dict[str, dict[Any, Any]]
        self._inventories_mtime: float | None
        self._inventories: list[str] = []
        self._inventory_error: str = ""
        self._runner: Command | AnsibleInventory

    @property
    def _inventory(self) -> dict[Any, Any]:
        """Return the inventory.

        Returns:
            The inventory details
        """
        return self.__inventory

    @_inventory.setter
    def _inventory(self, value: dict[Any, Any]) -> None:
        """Set the inventory and hostvars.

        Args:
            value: The inventory data
        """
        self.__inventory = value
        self._host_vars = {
            k: {**v, "inventory_hostname": k}
            for k, v in value.get("_meta", {}).get("hostvars", {}).items()
        }
        for group in self._inventory:
            for host in self._inventory[group].get("hosts", []):
                if host in self._host_vars:
                    continue
                self._host_vars[host] = {"inventory_hostname": host}

    @property
    def _show_columns(self) -> list[str]:
        """Return the columns to show for an inventory menu.

        Returns:
            The columns to show
        """
        if isinstance(self._args.inventory_column, list):
            return self._args.inventory_column
        return []

    def _set_inventories_mtime(self) -> None:
        """Record the inventory modification time."""
        modification_times = []
        for inventory in self._inventories:
            if Path(inventory).is_dir():
                modification_times.append(
                    max(Path(e).stat().st_mtime for e in Path(inventory).rglob("*")),
                )
            elif Path(inventory).is_file():
                modification_times.append(Path(inventory).stat().st_mtime)
        if modification_times:
            self._inventories_mtime = max(modification_times)
        else:
            self._inventories_mtime = None

    def update(self) -> None:
        """Request calling app update, inventory update checked in ``run()``."""
        self._calling_app.update()

    def _initialize_inventory(self) -> bool:
        """Set up args, collect inventory details, and check for errors.

        Returns:
            True if initialization succeeded, False otherwise
        """
        args_updated = self._update_args(
            [self._name] + shlex.split(self._interaction.action.match.groupdict()["params"] or ""),
        )
        if not args_updated:
            return False

        self.stdout = self._calling_app.stdout
        self._inventories = self._args.inventory
        self._set_inventories_mtime()

        self._collect_inventory_details()
        return bool(self._inventory) or bool(self._inventory_error)

    def _check_inventory_changed(self) -> bool:
        """Check if inventory files have been modified and refresh if needed.

        Returns:
            True if the loop should break, False otherwise
        """
        current_mtime = self._inventories_mtime
        self._set_inventories_mtime()
        if current_mtime != self._inventories_mtime:
            self._logger.debug("inventory changed")

            self._set_inventories_mtime()

            self._collect_inventory_details()
            if not self._inventory:
                return True

            if self._inventory_error:
                self._logger.error(self._inventory_error)
                return True

        return False

    def run(self, interaction: Interaction, app: AppPublic) -> Interaction | None:
        """Execute the ``inventory`` request for mode interactive.

        Args:
            interaction: The interaction from the user
            app: The app instance

        Returns:
            The pending
            :class:`~ansible_navigator.ui_framework.ui.Interaction` or
            :data:`None`
        """
        self._logger.debug("inventory requested in interactive mode")
        self._prepare_to_run(app, interaction)

        if not self._initialize_inventory():
            self._prepare_to_exit(interaction)
            return None

        if self._inventory_error:
            while True:
                interaction = self._interaction.ui.show(
                    self._inventory_error,
                    content_format=ContentFormat.ANSI,
                )
                if interaction.name != "refresh":
                    break
            self._prepare_to_exit(interaction)
            return None

        self.steps.append(self._build_main_menu())

        while True:
            self.update()
            self._take_step()
            if not self.steps:
                break

            if self._check_inventory_changed():
                break

            if self.steps.current.name == "quit":
                return self.steps.current

        self._prepare_to_exit(interaction)
        return None

    def run_stdout(self) -> RunStdoutReturn:
        """Execute the ``inventory`` request for mode stdout.

        Returns:
            The return code, 0 or 1. If the runner status if failed,
            return 1 along with a message to review the logs.
        """
        self._logger.debug("inventory requested in stdout mode")
        if hasattr(self._args, "inventory") and self._args.inventory:
            self._inventories = self._args.inventory
        self._collect_inventory_details()
        if self._runner.status == "failed":
            return RunStdoutReturn(message="Please review the log for errors.", return_code=1)
        return RunStdoutReturn(message="", return_code=0)

    def _take_step(self) -> None:
        """Take a step based on the current step or step back."""
        result = None
        if isinstance(self.steps.current, Interaction):
            result = run_action(self.steps.current.name, self.app, self.steps.current)
        elif isinstance(self.steps.current, Step):
            if self.steps.current.show_func:
                current_index = self.steps.current.index
                self.steps.current.show_func()
                self.steps.current.index = current_index

            if self.steps.current.type == "menu":
                result = self._interaction.ui.show(
                    obj=self.steps.current.value,
                    columns=self.steps.current.columns,
                    color_menu_item=color_menu,
                )
            elif self.steps.current.type == "content":
                result = self._interaction.ui.show(
                    obj=self.steps.current.value,
                    index=self.steps.current.index,
                    content_heading=content_heading,
                    filter_content_keys=filter_content_keys,
                )

        if result is None:
            self.steps.back_one()
        else:
            self.steps.append(result)

    def _build_main_menu(self) -> Step:
        """Build the inventory menu.

        Returns:
            The inventory menu definition
        """
        groups = MenuEntry(
            title="Browse groups",
            description="Explore each inventory group and group members members",
        )
        hosts = MenuEntry(
            title="Browse hosts",
            description="Explore the inventory with a list of all hosts",
        )

        step = Step(
            name="main_menu",
            columns=["title", "description"],
            select_func=self._step_from_main_menu,
            step_type="menu",
            value=[groups, hosts],
        )
        return step

    def _step_from_main_menu(self) -> Step:
        """Take a step away from the main menu.

        Raises:
            IndexError: When the index does not correspond to a menu
                entry

        Returns:
            Either the group or host menu
        """
        if self.steps.current.index == 0:
            return self._build_group_menu("all")
        if self.steps.current.index == 1:
            return self._build_host_menu()
        msg = "broken modules somewhere?"
        raise IndexError(msg)

    def _build_group_menu(self, key: str | None = None) -> Step:
        """Build the menu for inventory groups.

        Args:
            key: The optional menu name

        Returns:
            The inventory group menu definition
        """
        if key is None:
            key = self.steps.current.selected["__name"]

        try:
            menu = Menu()
            taxonomy = "\u25b8".join(
                ["all"]
                + [step.selected["__name"] for step in self.steps if step.name == "group_menu"],
            )

            columns = ["__name", "__taxonomy", "__type"]

            hosts = self._inventory[key].get("hosts", None)
            if hosts:
                columns.extend(self._show_columns)
                for host in sorted(hosts):
                    menu_entry = MenuEntry(**self._host_vars[host])
                    menu_entry["__name"] = menu_entry["inventory_hostname"]
                    menu_entry["__taxonomy"] = taxonomy
                    menu_entry["__type"] = "host"
                    menu.append(menu_entry)

            children = self._inventory[key].get("children", None)
            if children:
                for child in sorted(children):
                    menu_entry = MenuEntry()
                    menu_entry["__name"] = child
                    menu_entry["__taxonomy"] = taxonomy
                    menu_entry["__type"] = "group"
                    if hosts:
                        menu_entry.update(dict.fromkeys(self._show_columns, ""))
                    menu.append(menu_entry)

            return Step(
                name="group_menu",
                step_type="menu",
                value=menu,
                columns=columns,
                select_func=self._host_or_group_step,
                show_func=self._refresh,
            )
        except KeyError:
            # selected group was removed from inventory
            return self.steps.back_one()

    def _build_host_content(self) -> Step:
        """Build the inventory content for one host.

        Returns:
            The inventory content for the host
        """
        host_vars = self._host_vars
        try:
            values = [
                host_vars[m_entry.get("__name", m_entry.get("inventory_hostname"))]
                for m_entry in self.steps.current.value
                if "__type" not in m_entry or m_entry["__type"] == "host"
            ]
            entry = Step(
                name="host_content",
                step_type="content",
                value=values,
                index=self.steps.current.index,
                columns=["__name"] + self._show_columns,
                show_func=self._refresh,
            )
        except KeyError:
            # selected host removed from inventory
            return self.steps.back_one()
        return entry

    def _refresh(self) -> None:
        """Refresh the current step, removing and replacing it."""
        self.steps.back_one()
        self.steps.append(self.steps.current.select_func())

    def _build_host_menu(self) -> Step:
        """Build the menu of hosts.

        Returns:
            The hosts menu definition
        """
        menu = Menu()
        for host in self._host_vars.values():
            host["__type"] = "host"
            menu.append(MenuEntry(host))
        columns = ["inventory_hostname"] + self._show_columns
        return Step(
            columns=columns,
            name="host_menu",
            step_type="menu",
            value=menu,
            select_func=self._build_host_content,
            show_func=self._refresh,
        )

    def _host_or_group_step(self) -> Step:
        """Build a menu based on the type of the current step.

        Raises:
            TypeError: When the step type is unknown

        Returns:
            The group or host menu definition
        """
        if self.steps.current.selected["__type"] == "group":
            return self._build_group_menu()
        if self.steps.current.selected["__type"] == "host":
            return self._build_host_content()
        msg = "unknown step type"
        raise TypeError(msg)

    def _handle_inventory_errors(self, inventory_err: str, inventory_output: str) -> None:
        """Check inventory errors and show notifications or extract inventory.

        Args:
            inventory_err: The error output from the inventory collection
            inventory_output: The standard output from the inventory collection
        """
        preface = ["Errors were encountered while gathering the inventory:"]
        notify = ("ERROR!", "Error", "Unable to parse")
        if any(string in inventory_err for string in notify):
            if "[WARNING]" in inventory_err:
                parts = inventory_err.split("[WARNING]")
                messages = [part.replace("\n", " ").strip() for part in parts if part]
                present = preface + [f"[WARNING]{message}" for message in messages]
            else:
                present = preface + inventory_err.splitlines()
            for line in present:
                self._logger.error(line)
            warning = warning_notification(present)
            self._interaction.ui.show_form(warning)
        else:
            self._extract_inventory(inventory_output)

    def _collect_inventory_details_interactive(
        self,
        kwargs: dict[str, Any],
    ) -> None:
        """Use the runner subsystem to collect inventory details for mode interactive.

        Args:
            kwargs: The arguments for the runner call
        """
        try:
            # Extract the playbook dir the user may have provided
            index = self._args.cmdline.index("--playbook-dir")
            playbook_dir = self._args.cmdline[index + 1]
            source = "user provided"
        # Constants don't have an index, may go past end of list, param not found
        except (AttributeError, IndexError, ValueError):
            if isinstance(self._args.playbook, str):
                # or use the parent of the currently set playbook
                playbook_dir = f"{Path(self._args.playbook).resolve().parent}"
                source = "derived from playbook"
            else:
                # or the current working directory
                playbook_dir = str(Path.cwd())
                source = "CWD"
        self._logger.info("--playbook-directory for inventory from (%s): %s", source, playbook_dir)

        self._runner = AnsibleInventory(**kwargs)
        inventory_output, inventory_err = self._runner.fetch_inventory(
            action="list",
            inventories=self._inventories,
            playbook_dir=playbook_dir,
        )
        if inventory_output:
            parts = inventory_output.split("{", 1)
            inventory_err = parts[0] + inventory_err if inventory_err else parts[0]
            inventory_output = "{" + parts[1] if len(parts) == 2 else ""
        self._handle_inventory_errors(inventory_err, inventory_output)

    def _collect_inventory_details_automated(
        self,
        kwargs: dict[str, Any],
    ) -> tuple[str | None, str | None, int | None]:
        """Use the runner subsystem to collect inventory details for mode stdout.

        Args:
            kwargs: The arguments for the runner call

        Raises:
            RuntimeError: When the ``ansible-inventory`` command cannot
                be found for mode stdout

        Returns:
            The output, errors and return code from runner
        """
        if self._args.execution_environment:
            ansible_inventory_path = "ansible-inventory"
        else:
            exec_path = shutil.which("ansible-inventory")
            if exec_path is None:
                msg = "'ansible-inventory' executable not found"
                self._logger.error(msg)
                raise RuntimeError(msg)

            ansible_inventory_path = exec_path

        pass_through_arg = []
        if self._args.help_inventory is True:
            pass_through_arg.append("--help")

        if isinstance(self._args.cmdline, list):
            pass_through_arg.extend(self._args.cmdline)

        kwargs.update({"cmdline": pass_through_arg, "inventory": self._inventories})

        self._runner = Command(executable_cmd=ansible_inventory_path, **kwargs)
        stdout_return = self._runner.run()
        return stdout_return

    def _collect_inventory_details(
        self,
    ) -> tuple[str | None, str | None, int | None]:
        """Use the runner subsystem to collect inventory details for either mode.

        Returns:
            For mode interactive nothing. For mode stdout, the output,
            errors and return code from runner
        """
        if isinstance(self._args.set_environment_variable, dict):
            set_env_vars = {**self._args.set_environment_variable}
        else:
            set_env_vars = {}

        if self._args.display_color is False:
            set_env_vars["ANSIBLE_NOCOLOR"] = "1"

        kwargs = {
            "container_engine": self._args.container_engine,
            "host_cwd": str(Path.cwd()),
            "execution_environment_image": self._args.execution_environment_image,
            "execution_environment": self._args.execution_environment,
            "navigator_mode": self._args.mode,
            "pass_environment_variable": self._args.pass_environment_variable,
            "set_environment_variable": set_env_vars,
            "private_data_dir": self._args.ansible_runner_artifact_dir,
            "rotate_artifacts": self._args.ansible_runner_rotate_artifacts_count,
            "timeout": self._args.ansible_runner_timeout,
        }

        if isinstance(self._args.execution_environment_volume_mounts, list):
            kwargs.update(
                {"container_volume_mounts": self._args.execution_environment_volume_mounts},
            )

        if isinstance(self._args.container_options, list):
            kwargs.update({"container_options": self._args.container_options})

        if self._args.mode == "interactive":
            self._collect_inventory_details_interactive(kwargs)
        else:
            return self._collect_inventory_details_automated(kwargs)

        return (None, None, None)

    def _extract_inventory(self, stdout: str) -> None:
        """Load and the ``json`` output from the inventory collection process.

        Args:
            stdout: The output from the inventory collection process
        """
        try:
            self._inventory = json.loads(stdout)
        except json.JSONDecodeError as exc:
            self._logger.debug("json decode error: %s", str(exc))
            self._logger.debug("tried: %s", stdout)
            self._inventory_error = stdout



"""``:yaml`` command implementation."""

import logging

from ansible_navigator.app_public import AppPublic
from ansible_navigator.configuration_subsystem.definitions import ApplicationConfiguration
from ansible_navigator.content_defs import ContentFormat
from ansible_navigator.ui_framework import Interaction

from . import _actions as actions


@actions.register
class Action:
    """``:yaml`` command implementation."""

    KEGEX = r"^y(?:aml)?$"

    def __init__(self, args: ApplicationConfiguration) -> None:
        """Initialize the ``:yaml`` action.

        Args:
            args: The current settings for the application
        """
        self._args = args
        self._logger = logging.getLogger(__name__)

    def run(self, interaction: Interaction, app: AppPublic) -> None:
        """Execute the ``:yaml`` request for mode interactive.

        Args:
            interaction: The interaction from the user, action and value
            app: The app instance
        """
        self._logger.debug("yaml requested")
        if interaction.ui is not None:
            interaction.ui.scroll(0)
            content_format = interaction.ui.content_format(ContentFormat.YAML, default=True)
            self._logger.debug("Serialization set to %s", content_format)



"""One option in either a check or radio field."""

from __future__ import annotations

from dataclasses import dataclass

from .field_checks import FieldChecks
from .field_radio import FieldRadio


@dataclass
class FieldOption:
    """One option in either a check or radio field."""

    name: str
    text: str
    checked: bool = False
    disabled: bool = False

    def ansi_code(self, form_field: FieldChecks | FieldRadio) -> str:
        """Return our icon based on the form provided.

        Args:
            form_field: Form with check or radio field

        Raises:
            TypeError: If form_field is not a check or radio field

        Returns:
            Check box icon
        """
        if isinstance(form_field, FieldChecks):
            check_box = "\u25fc" if self.checked else "\u25fb"
            return check_box
        if isinstance(form_field, FieldRadio):
            check_box = "\u25cf" if self.checked else "\u25cb"
            return check_box
        raise TypeError



"""Abstractions for common serialization formats."""

from __future__ import annotations

import json
import logging
import os
import re
import tempfile

from dataclasses import is_dataclass
from functools import partial
from pathlib import Path
from typing import IO
from typing import Any
from typing import NamedTuple

import yaml

from ansible_navigator.content_defs import ContentBase
from ansible_navigator.content_defs import ContentFormat
from ansible_navigator.content_defs import ContentType
from ansible_navigator.content_defs import ContentView
from ansible_navigator.content_defs import SerializationFormat


logger = logging.getLogger(__name__)

UNKNOWN_SERIALIZATION_FORMAT_MSG = "Unknown serialization format"

try:
    from yaml import CSafeDumper as SafeDumper
except ImportError:
    from yaml import SafeDumper  # type: ignore

try:
    from yaml import CLoader as Loader
    from yaml import CSafeLoader as SafeLoader
except ImportError:
    from yaml import Loader  # type: ignore
    from yaml import SafeLoader  # type: ignore


def serialize(
    content: ContentType,
    content_view: ContentView,
    serialization_format: SerializationFormat,
) -> str | None:
    """Serialize a dataclass based on format and view.

    Args:
        content: The content dataclass to serialize
        content_view: The content view
        serialization_format: The serialization format

    Raises:
        ValueError: When serialization format is not recognized

    Returns:
        The serialized content
    """
    dumpable = _prepare_content(
        content=content,
        content_view=content_view,
        serialization_format=serialization_format,
    )
    if serialization_format == SerializationFormat.YAML:
        return _yaml_dumps(dumpable=dumpable)
    if serialization_format == SerializationFormat.JSON:
        return _json_dumps(dumpable=dumpable)
    raise ValueError(UNKNOWN_SERIALIZATION_FORMAT_MSG)


def serialize_write_file(
    content: ContentType,
    content_view: ContentView,
    file_mode: str,
    file: Path,
    serialization_format: SerializationFormat,
) -> None:
    """Serialize and write content to a file.

    Args:
        content: The content to serialize
        content_view: The content view
        file_mode: The file mode for the file
        file: The file to write to
        serialization_format: The serialization format

    Raises:
        ValueError: When serialization format is not recognized
    """
    dumpable = _prepare_content(
        content=content,
        content_view=content_view,
        serialization_format=serialization_format,
    )
    with file.open(mode=file_mode, encoding="utf-8") as file_handle:
        if serialization_format == SerializationFormat.JSON:
            _json_dump(dumpable=dumpable, file_handle=file_handle)
            return
        if serialization_format == SerializationFormat.YAML:
            _yaml_dump(dumpable=dumpable, file_handle=file_handle)
            return
    raise ValueError(UNKNOWN_SERIALIZATION_FORMAT_MSG)


def serialize_write_temp_file(
    content: ContentType,
    content_view: ContentView,
    content_format: ContentFormat,
) -> Path:
    """Serialize and write content to a premanent temporary file.

    Args:
        content: The content to serialize
        content_view: The content view
        content_format: The content format

    Raises:
        ValueError: When serialization format is not recognized

    Returns:
        A ``Path`` to the file written to
    """
    serialization_format = content_format.value.serialization
    suffix = content_format.value.file_extension

    if serialization_format is not None:
        dumpable = _prepare_content(
            content=content,
            content_view=content_view,
            serialization_format=serialization_format,
        )
    else:
        dumpable = content

    with tempfile.NamedTemporaryFile(suffix=suffix, delete=False, mode="w+t") as file_like:
        if content_format == ContentFormat.JSON:
            _json_dump(dumpable=dumpable, file_handle=file_like)
            return Path(file_like.name)
        if content_format == ContentFormat.YAML:
            _yaml_dump(dumpable=dumpable, file_handle=file_like)
            return Path(file_like.name)
        _text_dump(dumpable=str(dumpable), file_handle=file_like)
        return Path(file_like.name)
    raise ValueError(UNKNOWN_SERIALIZATION_FORMAT_MSG)


SERIALIZATION_FAILURE_MSG = (
    "The requested content could not be converted to {serialization_format}.\n"
    "The content was {content}\n"
    "Please log an issue for this, it should not have happened\n"
    "Error details: {exception_str}\n"
)


def _prepare_content(
    content: ContentType,
    content_view: ContentView,
    serialization_format: SerializationFormat,
) -> ContentType:
    """Prepare content for serialization.

    Args:
        content: The content to prepare
        content_view: The content view
        serialization_format: The serialization format

    Returns:
        The prepared content
    """
    if isinstance(content, tuple | list):
        if all(is_dataclass(c) for c in content):
            return [
                c.asdict(
                    content_view=content_view,
                    serialization_format=serialization_format,
                )
                for c in content
            ]
        return content
    if isinstance(content, bool | dict | float | int | str):
        return content
    if isinstance(content, ContentBase):
        return content.asdict(
            content_view=content_view,
            serialization_format=serialization_format,
        )

    # This is a big chance, it suggests all content is
    # one of ``ContentType``. the thinking is it's better
    # to error here early, then return a str(content)
    # Ideally this get caught in testing
    value_error = "Content type not recognized"
    error = SERIALIZATION_FAILURE_MSG.format(
        content=str(content),
        exception_str=value_error,
        serialization_format=serialization_format.value,
    )
    error += f"Content view: {content_view}\n"
    return error


class JsonParams(NamedTuple):
    """The parameters for json dump and dumps."""

    indent: int = 4
    sort_keys: bool = True
    ensure_ascii: bool = False


def _json_dump(dumpable: ContentType, file_handle: IO[str]) -> None:
    """Serialize the dumpable to json and write to a file.

    Args:
        dumpable: The object to dump
        file_handle: The file handle to write to
    """
    try:
        json.dump(dumpable, file_handle, **JsonParams()._asdict())
        file_handle.write("\n")  # Add newline json does not
    except TypeError as exc:
        error_message = SERIALIZATION_FAILURE_MSG.format(
            content=str(dumpable),
            exception_str=str(exc),
            serialization_format="JSON",
        )
        file_handle.write(error_message)
        logger.exception(error_message)


def _json_dumps(dumpable: ContentType) -> str:
    """Serialize the dumpable to json.

    Args:
        dumpable: The object to serialize

    Returns:
        The object serialized
    """
    try:
        return json.dumps(dumpable, **JsonParams()._asdict())
    except TypeError as exc:
        error_message = SERIALIZATION_FAILURE_MSG.format(
            content=str(dumpable),
            exception_str=str(exc),
            serialization_format="JSON",
        )
        logger.exception(error_message)
        return error_message


def _text_dump(dumpable: str, file_handle: IO[str]) -> None:
    """Write text to a file.

    Args:
        dumpable: The text to write
        file_handle: The file handle to write to
    """
    file_handle.write(dumpable)
    if not dumpable.endswith("\n"):
        file_handle.write("\n")


class YamlStyle(NamedTuple):
    """The parameters for yaml dump."""

    default_flow_style: bool = False
    explicit_start: bool = True
    allow_unicode: bool = True


def _yaml_dump(dumpable: ContentType, file_handle: IO[str]) -> None:
    """Serialize the dumpable to yaml and write to a file.

    Args:
        dumpable: The object to serialize
        file_handle: The file handle to write to
    """
    try:
        yaml.dump(dumpable, file_handle, Dumper=HumanDumper, **YamlStyle()._asdict())
    except yaml.representer.RepresenterError as exc:
        error_message = SERIALIZATION_FAILURE_MSG.format(
            content=str(dumpable),
            exception_str=str(exc),
            serialization_format="YAML",
        )
        file_handle.write(error_message)
        logger.exception(error_message)


def _yaml_dumps(dumpable: ContentType) -> str | None:
    """Serialize the dumpable to yaml.

    Args:
        dumpable: The object to serialize

    Returns:
        The serialized object
    """
    try:
        return yaml.dump(dumpable, Dumper=HumanDumper, **YamlStyle()._asdict())
    except yaml.representer.RepresenterError as exc:
        error_message = SERIALIZATION_FAILURE_MSG.format(
            content=str(dumpable),
            exception_str=str(exc),
            serialization_format="YAML",
        )
        logger.exception(error_message)
        return error_message


class HumanDumper(SafeDumper):
    # pylint: disable=too-many-ancestors
    """An instance of a pyyaml Dumper.

    This deviates from the base to dump a multiline string in a human readable format
    and disables the use of anchors and aliases.
    """

    def ignore_aliases(self, _data: Any) -> bool:
        """Disable the use of anchors and aliases in the given data.

        Args:
            _data: The data used to make the determination

        Returns:
            True, indicating aliases and anchors should not be used
        """
        return True

    def represent_scalar(
        self,
        tag: str,
        value: str,
        style: str | None = None,
    ) -> yaml.nodes.ScalarNode:
        """Represent all multiline strings as block scalars to improve readability for humans.

        Args:
            tag: A custom tag
            value: The value to represent
            style: The style to use

        Returns:
            The serialized multiline string, result of the super scalar
        """
        if style is None and _is_multiline_string(value):
            style = "|"

            # Remove leading or trailing newline and convert tabs to spaces
            # which can cause havoc on yaml blocks
            value = value.strip().expandtabs()

            # Replace some whitespace chars
            value = re.sub(r"\r", "", value)

        return super().represent_scalar(tag, value, style)


def _is_multiline_string(value: str) -> bool:
    """Determine if a string is multiline.

    Note:
        Inspired by http://stackoverflow.com/a/15423007/115478.

    Args:
        value: The value to check

    Returns:
        A boolean indicating if the string is multiline
    """
    return any(character in value for character in "\n\r\x1c\x1d\x1e\x85\u2028\u2029")


def opener(path: str, flags: int, mode: int) -> int:
    """Open a file with the given path and flags.

    Args:
        path: The path of the file to open.
        flags: The flags to use when opening the file.
        mode: The mode to use when opening the file.

    Returns:
        The file descriptor of the opened file.
    """
    return os.open(path=path, flags=flags, mode=mode)


def write_diagnostics_json(path: str, mode: int, content: object) -> None:
    """Write a file with the given name.

    Args:
        content: The content we want to write in the file.
        mode: The mode to use when writing the file.
        path: The path of the file to write.
    """
    # Without this, the created file will have 0o777 - 0o022 (default umask) = 0o755 permissions
    oldmask = os.umask(0)

    opener_func = partial(opener, mode=mode)
    with open(path, "w", encoding="utf-8", opener=opener_func) as f:
        f.write(json.dumps(content, indent=4, sort_keys=True))
    os.umask(oldmask)


__all__ = ("Loader", "SafeDumper", "SafeLoader", "SerializationFormat", "yaml")



"""Object definitions for the presentable transformation of the settings."""

from __future__ import annotations

from dataclasses import dataclass
from dataclasses import field
from typing import TYPE_CHECKING
from typing import Any
from typing import ClassVar
from typing import NewType
from typing import TypeVar

from ansible_navigator.content_defs import ContentBase

from .definitions import CliParameters
from .definitions import Constants as C
from .definitions import SettingsEntry
from .utils import create_settings_file_sample


if TYPE_CHECKING:
    from .navigator_configuration import Internals


PresentableSettingsEntryValue = bool | dict[Any, Any] | str | list[str]
CliT = TypeVar("CliT", bound="PresentableCliParameters")
EntT = TypeVar("EntT", bound="PresentableSettingsEntry")


@dataclass(frozen=True)
class PresentableCliParameters:
    """A settings entry's cli parameters in a presentable structure.

    Attributes:
        NO_LONG_MSG: Default message when no long CLI parameter exists.
        NO_SHORT_MSG: Default message when no short CLI parameter exists.
        long: The long cli parameter value.
        short: The short cli parameter value.
    """

    NO_LONG_MSG: ClassVar[str] = "No long CLI parameter"
    NO_SHORT_MSG: ClassVar[str] = "No short CLI parameter"

    long: str = NO_LONG_MSG
    short: str = NO_SHORT_MSG

    @classmethod
    def from_cli_params(
        cls,
        cli_parameters: CliParameters | None,
        name_dashed: str,
    ) -> PresentableCliParameters:
        """Create an ``_HRCliParameters`` based on an entry's cli parameters.

        Args:
            cli_parameters: The entry's cli parameters
            name_dashed: The dashed name of the parent settings entry

        Returns:
            The instance of self based on an entry's cli parameters
        """
        if isinstance(cli_parameters, CliParameters):
            short = cli_parameters.short or cls.NO_SHORT_MSG
            long = cli_parameters.long(name_dashed)
            return cls(long=long, short=short)
        return cls()


@dataclass
class PresentableSettingsEntry(ContentBase[Any]):
    # pylint: disable=too-many-instance-attributes
    """A settings entry in a presentable structure.

    Attributes:
        choices: The possible values.
        current_settings_file: The path to the current settings file.
        current_value: The current, effective value.
        default_value: The default value.
        default: Indicates if the current value == the default.
        description: A short description.
        env_var: The environment variable.
        name: The name.
        settings_file_sample: A sample settings file snippet.
        source: The source of the current value.
        subcommands: A list of subcommands where this entry is available.
        version_added: The version this entry was added in.
        cli_parameters: The cli parameters.
    """

    choices: list[str | bool]
    current_settings_file: str
    current_value: PresentableSettingsEntryValue
    default_value: PresentableSettingsEntryValue
    default: bool
    description: str
    env_var: str
    name: str
    settings_file_sample: str | dict[Any, Any]
    source: str
    subcommands: list[str]
    version_added: str
    cli_parameters: PresentableCliParameters = field(default_factory=PresentableCliParameters)

    @property
    def current(self) -> str:
        """Convert the current value into a string.

        Returns:
            The current value as a string
        """
        return str(self.current_value)

    def __lt__(self, other: object) -> bool:
        """Compare based on name, called by sort, sorted.

        Args:
            other: The entry to compare this to

        Returns:
            Indication of less than the other
        """
        if not isinstance(other, PresentableSettingsEntry):
            return NotImplemented
        return self.name < other.name

    @classmethod
    def for_settings_file(
        cls,
        all_subcommands: list[str],
        application_name: str,
        internals: Internals,
    ) -> PresentableSettingsEntry:
        """Create an ``PresentableSettingsEntry`` containing the details for the settings file.

        Args:
            all_subcommands: All application subcommands
            application_name: The application name
            internals: The internal storage for settings information

        Returns:
            The settings file entry
        """
        description = (
            "The path to the current settings file. Possible locations are"
            " {CWD}/ansible-navigator.{ext} or {HOME}/.ansible-navigator.{ext}"
            " where ext is yml, yaml or json."
        )
        return cls(
            choices=[],
            current_settings_file=internals.settings_file_path or C.NONE.value,
            current_value=internals.settings_file_path or C.NONE.value,
            default=internals.settings_source is C.NONE,
            default_value=C.NONE.value,
            description=description,
            name="Current settings file",
            env_var=f"{application_name.upper()}_CONFIG",
            settings_file_sample="Not applicable",
            source=internals.settings_source.value,
            subcommands=all_subcommands,
            version_added="v1.0",
        )

    @classmethod
    def from_settings_entry(
        cls,
        all_subcommands: list[str],
        application_name_dashed: str,
        entry: SettingsEntry,
        settings_file_path: str,
    ) -> PresentableSettingsEntry:
        """Create an ``PresentableSettingsEntry`` containing the details for one settings entry.

        Args:
            application_name_dashed: The application name, dashed
            all_subcommands: All application subcommands
            entry: A settings entry
            settings_file_path: The path to the settings file

        Returns:
            The settings file entry
        """
        cli_parameters = PresentableCliParameters.from_cli_params(
            cli_parameters=entry.cli_parameters,
            name_dashed=entry.name_dashed,
        )

        env_var = entry.environment_variable(application_name_dashed)
        entry_value_resolved = entry.value.resolved

        path = entry.settings_file_path(application_name_dashed)
        settings_file_sample = create_settings_file_sample(path, placeholder="<------")

        if isinstance(entry.subcommands, C):
            if entry.subcommands is C.ALL:
                subcommands = all_subcommands
            else:
                subcommands = [entry.subcommands.value]
        else:
            subcommands = entry.subcommands

        result = cls(
            choices=list(entry.choices),  # May be a tuple e.g. PLUGIN_TYPES
            cli_parameters=cli_parameters,
            current_settings_file=str(settings_file_path),
            current_value=entry_value_resolved.current,
            default=entry_value_resolved.is_default,
            default_value=entry_value_resolved.default,
            description=entry.short_description,
            env_var=env_var,
            name=entry.name.replace("_", " ").capitalize(),
            settings_file_sample=settings_file_sample,
            source=entry.value.source.value,
            subcommands=subcommands,
            version_added=entry.version_added,
        )
        return result


PresentableSettingsEntries = NewType(
    "PresentableSettingsEntries",
    tuple[PresentableSettingsEntry, ...],
)



"""Herein lies the ability for ansible-runner to run the ansible-doc command."""

from __future__ import annotations

from typing import Any

from ansible_runner import get_plugin_docs

from .base import Base


class AnsibleDoc(Base):
    """An interface to ``ansible-runner`` for running the ``ansible-doc`` command."""

    def fetch_plugin_doc(
        self,
        plugin_names: list[str | None],
        plugin_type: str | None = None,
        response_format: str | None = "json",
        snippet: bool | None = None,
        playbook_dir: str | None = None,
        module_path: str | None = None,
    ) -> tuple[dict[Any, Any] | str, dict[Any, Any] | str]:
        """Run ``ansible-doc`` command and get the plugin docs related details.

        Args:
            plugin_names: The name of the plugins to get docs
            plugin_type: The type of the plugin mentioned in
                plugins_names Valid values are ``become``, ``cache``,
                ``callback``, ``cliconf``, ``connection``, ``httpapi``,
                ``inventory``, ``lookup``, ``netconf``, ``shell``,
                ``vars``, ``module``, ``strategy``. If the value is not
                provided it defaults to ``module``.
            response_format: The output format for response. Valid
                values can be one of ``json`` or ``human`` and the
                response is either json string or plain text in human
                readable format. Defaults to ``json``.
            snippet: Show playbook snippet for specified plugin(s)
            playbook_dir: This parameter is used to sets the relative
                path to handle playbook adjacent installed plugins
            module_path: This parameter is prepend colon-separated
                path(s) to module library
                (default=~/.ansible/plugins/modules:
                /usr/share/ansible/plugins/modules).

        Returns:
            A tuple of response and error string. If the value of
            ``response_format`` is ``json`` it returns a python
            dictionary object.
        """
        return get_plugin_docs(
            plugin_names,
            plugin_type=plugin_type,
            response_format=response_format,
            snippet=snippet,
            playbook_dir=playbook_dir,
            module_path=module_path,
            **self._runner_args,
        )



"""An interface to use a sqlite database as a key-value store."""

from __future__ import annotations

import sqlite3

from collections.abc import ItemsView
from collections.abc import Iterator
from collections.abc import KeysView
from collections.abc import MutableMapping
from collections.abc import ValuesView
from typing import TYPE_CHECKING


if TYPE_CHECKING:
    from pathlib import Path


class KVSKeysView(KeysView[str]):
    """A glorified KeysView specific to, and returned by, methods in KeyValueStore."""


class KVSItemsView(ItemsView[str, str]):
    """A glorified ItemsView specific to, and returned by, methods in KeyValueStore."""


class KVSValuesView(ValuesView[str]):
    """A glorified ValuesView specific to, and returned by, methods in KeyValueStore."""


class KeyValueStore(MutableMapping[str, str]):
    """An interface to use a sqlite database as a key-value store."""

    def __init__(self, filename: str | Path) -> None:
        """Initialize the key-value store.

        Args:
            filename: The full path to the sqlite database file
        """
        self._path = str(filename)
        self.conn = sqlite3.connect(self._path)
        cursor = self.conn.cursor()
        cursor.execute("CREATE TABLE IF NOT EXISTS kv (key text unique, value text)")

    @property
    def path(self) -> str:
        """Provide the filename where the KVS is stored on disk.

        Returns:
            The path to the key-value store
        """
        return self._path

    def close(self) -> None:
        """Close the connection to the database."""
        self.conn.commit()
        self.conn.close()

    def open_(self) -> sqlite3.Connection:
        """Establish the connection to the database.

        Returns:
            A connection to the database
        """
        self.conn = sqlite3.connect(self.path)
        return self.conn

    def __len__(self) -> int:
        """Count the number of keys in the key-value store.

        Returns:
            The number of keys

        Raises:
            TypeError: If there is a typing error.
        """
        cursor = self.conn.cursor()
        rows = cursor.execute("SELECT COUNT(*) FROM kv").fetchone()[0]
        if not isinstance(rows, int):
            raise TypeError
        return rows if rows is not None else 0

    def iterkeys(self) -> Iterator[str]:
        """Yield keys from the key-value store one by one.

        Yields:
            The keys in the key-value store
        """
        cursor = self.conn.cursor()
        for row in cursor.execute("SELECT key FROM kv"):
            yield row[0]

    def itervalues(self) -> Iterator[str]:
        """Yield values from the key-value store one by one.

        Yields:
            The values in the key-value store
        """
        cursor = self.conn.cursor()
        for row in cursor.execute("SELECT value FROM kv"):
            yield row[0]

    def iteritems(self) -> Iterator[tuple[str, str]]:
        """Yield items from the key-value store one by one.

        Yields:
            The key-value store as items (key, value)
        """
        cursor = self.conn.cursor()
        for row in cursor.execute("SELECT key, value FROM kv"):
            yield row[0], row[1]

    def keys(self) -> KVSKeysView:
        """Return all keys in the key-value store.

        Returns:
            All the keys
        """
        return KVSKeysView(self)

    def values(self) -> KVSValuesView:
        """Return all values in the key-value store.

        Returns:
            All the values
        """
        return KVSValuesView(self)

    def items(self) -> KVSItemsView:
        """Return all items from the key-value store.

        Returns:
            The key-value store as items (key, value)
        """
        return KVSItemsView(self)

    # mypy complains about this 'str' because it's more specific than the 'object'
    # that dict's Mapping superclass wants. However, it's correct in our case.
    # Our sqlite table expects string keys (column type 'text'). It's an error to
    # pass something else. If we pass something too weird, the sqlite3 lib will
    # throw an error at runtime.
    def __contains__(self, key: str) -> bool:  # type: ignore[override]
        """Check if a given key is in the key-value store.

        This provides dictionary like  `some_key in` support for the key-value store.

        Args:
            key: The key to search for

        Returns:
            An indication of the provided key's existence in the key-
            value store
        """
        cursor = self.conn.cursor()
        return cursor.execute("SELECT 1 FROM kv WHERE key = ?", (key,)).fetchone() is not None

    def __getitem__(self, key: str) -> str:
        """Return a value from the key-value store given a key.

        Args:
            key: The key to find

        Raises:
            KeyError: When the key does not exist in the key-value store

        Returns:
            The value for the key provided
        """
        cursor = self.conn.cursor()
        item = cursor.execute("SELECT value FROM kv WHERE key = ?", (key,)).fetchone()
        if item is None:
            raise KeyError(key)
        return item[0]

    def __setitem__(self, key: str, value: str) -> None:
        """Place a key-value combination in the key-value store.

        Args:
            key: The key of the combination to set
            value: The value of the combination to set
        """
        cursor = self.conn.cursor()
        cursor.execute("REPLACE INTO kv (key, value) VALUES (?,?)", (key, value))

    def __delitem__(self, key: str) -> None:
        """Delete a key-value combination in the key-value store.

        Args:
            key: The key of the entry to delete

        Raises:
            KeyError: When the provided key does not exist in the key-
                value store
        """
        if key not in self:
            raise KeyError(key)
        cursor = self.conn.cursor()
        cursor.execute("DELETE FROM kv WHERE key = ?", (key,))

    def __iter__(self) -> Iterator[str]:
        """Yield values extracted from the key-value store one by one.

        This provides dictionary like `for key in` support for the key-value store.

        Returns:
            The keys in the key-value store
        """
        return self.iterkeys()

    def __repr__(self) -> str:
        """Represent the key-value store.

        Returns:
            Representation of the key-value store
        """
        # Loosely based on collections.OrderedDict#__repr__
        if not self:
            return f"{self.__class__.__name__}()"
        return f"{self.__class__.__name__}({list(self.items())})"



"""``:sample_working`` command implementation."""

import time

from ansible_navigator.action_base import ActionBase
from ansible_navigator.app_public import AppPublic
from ansible_navigator.configuration_subsystem.definitions import ApplicationConfiguration
from ansible_navigator.ui_framework import Interaction
from ansible_navigator.ui_framework import nonblocking_notification

from . import _actions as actions


@actions.register
class Action(ActionBase):
    """``:sample_working`` command implementation."""

    KEGEX = r"^sample_working$"

    def __init__(self, args: ApplicationConfiguration) -> None:
        """Initialize the ``:sample_working`` action.

        Args:
            args: The current settings for the application
        """
        super().__init__(args=args, logger_name=__name__, name="sample_working")

    def run(self, interaction: Interaction, app: AppPublic) -> None:
        """Execute the ``:sample_working`` request for mode interactive.

        Args:
            interaction: The interaction from the user
            app: The app instance
        """
        self._logger.debug("sample working requested")
        self._prepare_to_run(app, interaction)
        messages = ["Please wait, this won't take long, about 3 seconds, really."]

        form = nonblocking_notification(messages)
        interaction.ui.show_form(form)
        time.sleep(3)

        self._prepare_to_exit(interaction)
